from openfisca_france.model.base import *

# Variables apparaissant dans la feuille de déclaration de patrimoine soumis à l'ISF

# Immeubles bâtis


class b1ab(Variable):
    value_type = int
    unit = 'currency'
    entity = FoyerFiscal
    label = 'Valeur de la résidence principale avant abattement'
    definition_period = YEAR


class b1ac(Variable):
    value_type = int
    unit = 'currency'
    entity = FoyerFiscal
    label = 'Valeur des autres immeubles avant abattement'
    definition_period = YEAR


# non bâtis
class b1bc(Variable):
    value_type = int
    unit = 'currency'
    entity = FoyerFiscal
    label = 'Immeubles non bâtis : bois, fôrets et parts de groupements forestiers'
    definition_period = YEAR


class b1be(Variable):
    value_type = int
    unit = 'currency'
    entity = FoyerFiscal
    label = 'Immeubles non bâtis : biens ruraux loués à long termes'
    definition_period = YEAR


class b1bh(Variable):
    value_type = int
    unit = 'currency'
    entity = FoyerFiscal
    label = 'Immeubles non bâtis : parts de groupements fonciers agricoles et de groupements agricoles fonciers'
    definition_period = YEAR


class b1bk(Variable):
    value_type = int
    unit = 'currency'
    entity = FoyerFiscal
    label = 'Immeubles non bâtis : autres biens'
    definition_period = YEAR


# droits sociaux- valeurs mobilières-liquidités- autres meubles
class b1cl(Variable):
    value_type = int
    unit = 'currency'
    entity = FoyerFiscal
    label = 'Parts et actions détenues par les salariés et mandataires sociaux'
    definition_period = YEAR


class b1cb(Variable):
    value_type = int
    unit = 'currency'
    entity = FoyerFiscal
    label = 'Parts et actions de sociétés avec engagement de conservation de 6 ans minimum'
    definition_period = YEAR


class b1cd(Variable):
    value_type = int
    unit = 'currency'
    entity = FoyerFiscal
    label = 'Droits sociaux de sociétés dans lesquelles vous exercez une fonction ou une activité'
    definition_period = YEAR


class b1ce(Variable):
    value_type = int
    unit = 'currency'
    entity = FoyerFiscal
    label = 'Autres valeurs mobilières'
    definition_period = YEAR


class b1cf(Variable):
    value_type = int
    unit = 'currency'
    entity = FoyerFiscal
    label = 'Liquidités'
    definition_period = YEAR


class b1cg(Variable):
    value_type = int
    unit = 'currency'
    entity = FoyerFiscal
    label = 'Autres biens meubles'
    definition_period = YEAR


class b1co(Variable):
    value_type = int
    unit = 'currency'
    entity = FoyerFiscal
    label = "Autres biens meubles : contrats d'assurance-vie"
    definition_period = YEAR


#    b1ch
#    b1ci
#    b1cj
#    b1ck


# passifs et autres réductions
class b2gh(Variable):
    value_type = int
    unit = 'currency'
    entity = FoyerFiscal
    label = 'Total du passif et autres déductions'
    definition_period = YEAR


# réductions
class b2mt(Variable):
    value_type = int
    unit = 'currency'
    entity = FoyerFiscal
    label = 'Réductions pour investissements directs dans une société'
    definition_period = YEAR


class b2ne(Variable):
    value_type = int
    unit = 'currency'
    entity = FoyerFiscal
    label = 'Réductions pour investissements directs dans une société'
    definition_period = YEAR


class b2mv(Variable):
    value_type = int
    unit = 'currency'
    entity = FoyerFiscal
    label = 'Réductions pour investissements par sociétés interposées, holdings'
    definition_period = YEAR


class b2nf(Variable):
    value_type = int
    unit = 'currency'
    entity = FoyerFiscal
    label = 'Réductions pour investissements par sociétés interposées, holdings'
    definition_period = YEAR


class b2mx(Variable):
    value_type = int
    unit = 'currency'
    entity = FoyerFiscal
    label = 'Réductions pour investissements par le biais de FIP'
    definition_period = YEAR


class b2na(Variable):
    value_type = int
    unit = 'currency'
    entity = FoyerFiscal
    label = 'Réductions pour investissements par le biais de FCPI ou FCPR'
    definition_period = YEAR


class b2nc(Variable):
    value_type = int
    unit = 'currency'
    entity = FoyerFiscal
    label = "Réductions pour dons à certains organismes d'intérêt général"
    definition_period = YEAR


# montant impôt acquitté hors de France
class b4rs(Variable):
    value_type = int
    unit = 'currency'
    entity = FoyerFiscal
    label = "Montant de l'impôt acquitté hors de France"
    definition_period = YEAR


# BOUCLIER FISCAL

class rev_or(Variable):
    value_type = int
    unit = 'currency'
    entity = FoyerFiscal
    definition_period = YEAR


class rev_exo(Variable):
    value_type = int
    unit = 'currency'
    entity = FoyerFiscal
    definition_period = YEAR


class tax_fonc(Variable):
    value_type = int
    unit = 'currency'
    entity = FoyerFiscal
    label = 'Taxe foncière'
    definition_period = YEAR


class restit_imp(Variable):
    value_type = int
    unit = 'currency'
    entity = FoyerFiscal
    definition_period = YEAR


class etr(Variable):
    value_type = int
    entity = Individu
    definition_period = MONTH
    set_input = set_input_divide_by_period


# Calcul de l'impôt de solidarité sur la fortune

# 1 ACTIF BRUT

class isf_ifi_imm_bati(Variable):
    value_type = float
    entity = FoyerFiscal
    label = "Base de l'ISF-IFI sur l'immobilier bâti"
    definition_period = YEAR

    def formula_2018_01_01(foyer_fiscal, period, parameters):
        b1ab = foyer_fiscal('b1ab', period)
        b1ac = foyer_fiscal('b1ac', period)
        reduc_exo = parameters(period).taxation_capital.impot_fortune_immobiliere_ifi_partir_2018.reduc_exo

        return (1 - reduc_exo.abattement_residence_principale) * b1ab + b1ac

    def formula(foyer_fiscal, period, parameters):
        b1ab = foyer_fiscal('b1ab', period)
        b1ac = foyer_fiscal('b1ac', period)
        reduc_exo = parameters(period).taxation_capital.impot_solidarite_fortune_isf_1989_2017.reduc_exo

        return (1 - reduc_exo.abattement_residence_principale) * b1ab + b1ac


class isf_ifi_imm_non_bati(Variable):
    value_type = float
    entity = FoyerFiscal
    label = "Base de l'ISF-IFI sur l'immobilier non-bâti"
    definition_period = YEAR

    def formula_2018_01_01(foyer_fiscal, period, parameters):
        b1bc = foyer_fiscal('b1bc', period)
        b1be = foyer_fiscal('b1be', period)
        b1bh = foyer_fiscal('b1bh', period)
        b1bk = foyer_fiscal('b1bk', period)
        non_bati = parameters(period).taxation_capital.impot_fortune_immobiliere_ifi_partir_2018.forfait_mobilier.non_bati

        # forêts
        b1bd = b1bc * non_bati.taux_bois_forets
        # bien ruraux loués à long terme
        b1bf = min_(b1be, non_bati.seuil) * non_bati.taux_biens_ruraux
        b1bg = max_(b1be - non_bati.seuil, 0) * non_bati.taux_forestier_agricole
        # part de groupements forestiers- agricoles fonciers
        b1bi = min_(b1bh, non_bati.seuil) * non_bati.taux_biens_ruraux
        b1bj = max_(b1bh - non_bati.seuil, 0) * non_bati.taux_forestier_agricole
        return b1bd + b1bf + b1bg + b1bi + b1bj + b1bk

    def formula(foyer_fiscal, period, parameters):
        b1bc = foyer_fiscal('b1bc', period)
        b1be = foyer_fiscal('b1be', period)
        b1bh = foyer_fiscal('b1bh', period)
        b1bk = foyer_fiscal('b1bk', period)
        non_bati = parameters(period).taxation_capital.impot_solidarite_fortune_isf_1989_2017.forfait_mobilier.non_bati

        # forêts
        b1bd = b1bc * non_bati.taux_bois_forets
        # bien ruraux loués à long terme
        b1bf = min_(b1be, non_bati.seuil) * non_bati.taux_biens_ruraux
        b1bg = max_(b1be - non_bati.seuil, 0) * non_bati.taux_forestier_agricole
        # part de groupements forestiers- agricoles fonciers
        b1bi = min_(b1bh, non_bati.seuil) * non_bati.taux_biens_ruraux
        b1bj = max_(b1bh - non_bati.seuil, 0) * non_bati.taux_forestier_agricole
        return b1bd + b1bf + b1bg + b1bi + b1bj + b1bk

# droits sociaux- valeurs mobilières- liquidités- autres meubles


class isf_actions_sal(Variable):  # non présent en 2005
    value_type = float
    entity = FoyerFiscal
    label = 'isf_actions_sal'
    definition_period = YEAR
    end = '2017-12-31'

    def formula_2006(foyer_fiscal, period, parameters):
        '''
        Parts ou actions détenues par les salariés et mandataires sociaux
        '''
        b1cl = foyer_fiscal('b1cl', period)
        droits_sociaux = parameters(period).taxation_capital.impot_solidarite_fortune_isf_1989_2017.droits_sociaux

        return b1cl * droits_sociaux.taux_salaries_mandataires_sociaux


class isf_droits_sociaux(Variable):
    value_type = float
    entity = FoyerFiscal
    label = 'isf_droits_sociaux'
    definition_period = YEAR
    end = '2017-12-31'

    def formula(foyer_fiscal, period, parameters):
        isf_actions_sal = foyer_fiscal('isf_actions_sal', period)
        b1cb = foyer_fiscal('b1cb', period)
        b1cd = foyer_fiscal('b1cd', period)
        b1ce = foyer_fiscal('b1ce', period)
        b1cf = foyer_fiscal('b1cf', period)
        b1cg = foyer_fiscal('b1cg', period)
        droits_sociaux = parameters(period).taxation_capital.impot_solidarite_fortune_isf_1989_2017.droits_sociaux

        b1cc = b1cb * droits_sociaux.taux_actions_conservees_6_ans
        return isf_actions_sal + b1cc + b1cd + b1ce + b1cf + b1cg


class assiette_isf_ifi(Variable):
    value_type = float
    entity = FoyerFiscal
    label = "Assiette de l'ISF-IFI"
    definition_period = YEAR

    def formula_2018_01_01(foyer_fiscal, period, parameters):
        # TODO: Gérer les trois option meubles meublants
        isf_ifi_imm_bati = foyer_fiscal('isf_ifi_imm_bati', period)
        isf_ifi_imm_non_bati = foyer_fiscal('isf_ifi_imm_non_bati', period)
        isf_droits_sociaux = foyer_fiscal('isf_droits_sociaux', period)
        b1cg = foyer_fiscal('b1cg', period)
        b2gh = foyer_fiscal('b2gh', period)
        forfait_mobilier = parameters(period).taxation_capital.impot_fortune_immobiliere_ifi_partir_2018.forfait_mobilier

        total = isf_ifi_imm_bati + isf_ifi_imm_non_bati + isf_droits_sociaux
        forf_mob = (b1cg != 0) * b1cg + (b1cg == 0) * total * forfait_mobilier.majoration_forfaitaire
        actif_brut = total + forf_mob
        return actif_brut - b2gh

    def formula_1989_01_01(foyer_fiscal, period, parameters):
        # TODO: Gérer les trois option meubles meublants
        isf_ifi_imm_bati = foyer_fiscal('isf_ifi_imm_bati', period)
        isf_ifi_imm_non_bati = foyer_fiscal('isf_ifi_imm_non_bati', period)
        isf_droits_sociaux = foyer_fiscal('isf_droits_sociaux', period)
        b1cg = foyer_fiscal('b1cg', period)
        b2gh = foyer_fiscal('b2gh', period)
        forfait_mobilier = parameters(period).taxation_capital.impot_solidarite_fortune_isf_1989_2017.forfait_mobilier

        total = isf_ifi_imm_bati + isf_ifi_imm_non_bati + isf_droits_sociaux
        forf_mob = (b1cg != 0) * b1cg + (b1cg == 0) * total * forfait_mobilier.majoration_forfaitaire
        actif_brut = total + forf_mob
        return actif_brut - b2gh

    def formula(foyer_fiscal, period, parameters):
        # TODO: Gérer les trois option meubles meublants
        isf_ifi_imm_bati = foyer_fiscal('isf_ifi_imm_bati', period)
        isf_ifi_imm_non_bati = foyer_fiscal('isf_ifi_imm_non_bati', period)
        isf_droits_sociaux = foyer_fiscal('isf_droits_sociaux', period)
        b1cg = foyer_fiscal('b1cg', period)
        b2gh = foyer_fiscal('b2gh', period)
        forfait_mobilier = parameters(period).taxation_capital.impot_grandes_fortunes_1982_1986.forfait_mobilier

        total = isf_ifi_imm_bati + isf_ifi_imm_non_bati + isf_droits_sociaux
        forf_mob = (b1cg != 0) * b1cg + (b1cg == 0) * total * forfait_mobilier.majoration_forfaitaire
        actif_brut = total + forf_mob
        return actif_brut - b2gh

# Calcul de l'impôt par application du barème


class isf_ifi_iai(Variable):
    value_type = float
    entity = FoyerFiscal
    label = 'ISF-IFI avant décote, réductions et plafonnement'
    definition_period = YEAR

    def formula_2018_01_01(foyer_fiscal, period, parameters):
        assiette_isf_ifi = foyer_fiscal('assiette_isf_ifi', period)
        bareme = parameters(period).taxation_capital.impot_fortune_immobiliere_ifi_partir_2018.bareme.bareme
        return bareme.calc(assiette_isf_ifi)

    # TODO: Cette formule a seulement été vérifiée jusqu'au 2015-12-31
    def formula_2011_01_01(foyer_fiscal, period, parameters):
        assiette_isf_ifi = foyer_fiscal('assiette_isf_ifi', period)
        bareme = parameters(period).taxation_capital.impot_solidarite_fortune_isf_1989_2017.bareme.bareme
        assiette_isf_ifi = (assiette_isf_ifi >= bareme.rates[1]) * assiette_isf_ifi
        return bareme.calc(assiette_isf_ifi)

    def formula_2002_01_01(foyer_fiscal, period, parameters):
        assiette_isf_ifi = foyer_fiscal('assiette_isf_ifi', period)
        bareme = parameters(period).taxation_capital.impot_solidarite_fortune_isf_1989_2017.bareme.bareme
        return bareme.calc(assiette_isf_ifi)


class isf_ifi_avant_reduction(Variable):
    value_type = float
    entity = FoyerFiscal
    label = 'ISF-IFI avant réductions et plafonnement'
    definition_period = YEAR

    def formula(foyer_fiscal, period, parameters):
        isf_ifi_iai = foyer_fiscal('isf_ifi_iai', period)
        decote_isf_ifi = foyer_fiscal('decote_isf_ifi', period)

        return isf_ifi_iai - decote_isf_ifi


class isf_reduc_pac(Variable):
    value_type = float
    entity = FoyerFiscal
    label = 'isf_reduc_pac'
    end = '2012-12-31'
    definition_period = YEAR

    def formula(foyer_fiscal, period, parameters):
        '''
        Réductions pour personnes à charges
        '''
        nb_pac = foyer_fiscal('nb_pac', period)
        nbH = foyer_fiscal('nbH', period)
        reduc_exo = parameters(period).taxation_capital.impot_solidarite_fortune_isf_1989_2017.reduc_exo

        return reduc_exo.reduction_enfant_charge * nb_pac + (reduc_exo.reduction_enfant_charge / 2) * nbH


class isf_inv_pme(Variable):
    value_type = float
    entity = FoyerFiscal
    label = 'isf_inv_pme'
    definition_period = YEAR

    def formula_2008(foyer_fiscal, period, parameters):
        '''
        Réductions pour investissements dans les PME
        à partir de 2008!
        '''
        b2mt = foyer_fiscal('b2mt', period)
        b2ne = foyer_fiscal('b2ne', period)
        b2mv = foyer_fiscal('b2mv', period)
        b2nf = foyer_fiscal('b2nf', period)
        b2mx = foyer_fiscal('b2mx', period)
        b2na = foyer_fiscal('b2na', period)
        reduc_impot = parameters(period).taxation_capital.impot_solidarite_fortune_isf_1989_2017.reduc_impot
        taux_dons = reduc_impot.reduction_dons_certains_organismes_interet_general.taux
        taux_invest_direct = reduc_impot.reduction_investissements_capital_pme.taux_investissement_direct
        taux_fip_fci = reduc_impot.reduction_investissements_dans_fcpi_ou_fip_dans_pme.taux_investissement

        inv_dir_soc = b2mt * taux_dons + b2ne * taux_invest_direct
        holdings = b2mv * taux_dons + b2nf * taux_invest_direct
        fip = b2mx * taux_fip_fci
        fcpi = b2na * taux_fip_fci

        montant_reduc = holdings + fip + fcpi + inv_dir_soc
        plaf = parameters(period).taxation_capital.impot_solidarite_fortune_isf_1989_2017.reduc_impot.plafond_somme_trois_reductions_pme_fcip_fip_pme_dons

        return where(montant_reduc < plaf, montant_reduc, plaf)

    def formula_2018(foyer_fiscal, period, parameters):
        '''
        Réductions pour investissements dans les PME
        à partir de 2008!
        '''
        b2mt = foyer_fiscal('b2mt', period)
        b2ne = foyer_fiscal('b2ne', period)
        b2mv = foyer_fiscal('b2mv', period)
        b2nf = foyer_fiscal('b2nf', period)
        b2mx = foyer_fiscal('b2mx', period)
        b2na = foyer_fiscal('b2na', period)
        reduc_impot = parameters(period).taxation_capital.impot_fortune_immobiliere_ifi_partir_2018.reduc_impot
        taux_dons = reduc_impot.reduction_dons_certains_organismes_interet_general.taux
        taux_invest_direct = reduc_impot.reduction_investissements_capital_pme.taux_investissement_direct
        taux_fip_fci = reduc_impot.reduction_investissements_dans_fcpi_ou_fip_dans_pme.taux_investissement

        inv_dir_soc = b2mt * taux_dons + b2ne * taux_invest_direct
        holdings = b2mv * taux_dons + b2nf * taux_invest_direct
        fip = b2mx * taux_fip_fci
        fcpi = b2na * taux_fip_fci

        montant_reduc = holdings + fip + fcpi + inv_dir_soc
        plaf = parameters(period).taxation_capital.impot_fortune_immobiliere_ifi_partir_2018.reduc_impot.plafond_somme_trois_reductions_pme_fcip_fip_pme_dons

        return where(montant_reduc < plaf, montant_reduc, plaf)


class isf_org_int_gen(Variable):
    value_type = float
    entity = FoyerFiscal
    label = 'isf_org_int_gen'
    definition_period = YEAR

    def formula_2008(foyer_fiscal, period, parameters):
        b2nc = foyer_fiscal('b2nc', period)
        P = parameters(period).taxation_capital.impot_solidarite_fortune_isf_1989_2017.reduc_impot.reduction_dons_certains_organismes_interet_general
        montant = b2nc * P.taux

        return where(montant < P.limite_reduction, montant, P.limite_reduction)

    def formula_2018(foyer_fiscal, period, parameters):
        b2nc = foyer_fiscal('b2nc', period)
        P = parameters(period).taxation_capital.impot_fortune_immobiliere_ifi_partir_2018.reduc_impot.reduction_dons_certains_organismes_interet_general
        montant = b2nc * P.taux

        return where(montant < P.limite_reduction, montant, P.limite_reduction)


class isf_ifi_avant_plaf(Variable):
    value_type = float
    entity = FoyerFiscal
    label = 'ISF-IFI avant plafonnement'
    definition_period = YEAR

    def formula(foyer_fiscal, period, parameters):
        isf_ifi_avant_reduction = foyer_fiscal('isf_ifi_avant_reduction', period)
        isf_reduc_pac = foyer_fiscal('isf_reduc_pac', period)

        return max_(0, isf_ifi_avant_reduction - isf_reduc_pac)

    def formula_2008(foyer_fiscal, period, parameters):
        isf_ifi_avant_reduction = foyer_fiscal('isf_ifi_avant_reduction', period)
        isf_inv_pme = foyer_fiscal('isf_inv_pme', period)
        isf_org_int_gen = foyer_fiscal('isf_org_int_gen', period)
        isf_reduc_pac = foyer_fiscal('isf_reduc_pac', period)

        return max_(0, isf_ifi_avant_reduction - (isf_inv_pme + isf_org_int_gen) - isf_reduc_pac)

    def formula_2009(foyer_fiscal, period, parameters):
        isf_ifi_avant_reduction = foyer_fiscal('isf_ifi_avant_reduction', period)
        isf_inv_pme = foyer_fiscal('isf_inv_pme', period)
        isf_org_int_gen = foyer_fiscal('isf_org_int_gen', period)
        isf_reduc_pac = foyer_fiscal('isf_reduc_pac', period)
        borne_max = parameters(period).taxation_capital.impot_solidarite_fortune_isf_1989_2017.reduc_impot.plafond_somme_trois_reductions_pme_fcip_fip_pme_dons

        return max_(0, isf_ifi_avant_reduction - min_(isf_inv_pme + isf_org_int_gen, borne_max) - isf_reduc_pac)

    def formula_2018(foyer_fiscal, period, parameters):
        isf_ifi_avant_reduction = foyer_fiscal('isf_ifi_avant_reduction', period)
        isf_inv_pme = foyer_fiscal('isf_inv_pme', period)
        isf_org_int_gen = foyer_fiscal('isf_org_int_gen', period)
        isf_reduc_pac = foyer_fiscal('isf_reduc_pac', period)
        borne_max = parameters(period).taxation_capital.impot_fortune_immobiliere_ifi_partir_2018.reduc_impot.plafond_somme_trois_reductions_pme_fcip_fip_pme_dons

        return max_(0, isf_ifi_avant_reduction - min_(isf_inv_pme + isf_org_int_gen, borne_max) - isf_reduc_pac)


# # calcul du plafonnement ##
class total_impots_plafonnement_isf_ifi(Variable):
    value_type = float
    entity = FoyerFiscal
    label = "Total des impôts dus au titre des revenus et produits (irpp, contribution_exceptionnelle_hauts_revenus, prélèvements forfaitaires, prélèvements sociaux) + ISF et IFI. Utilisé pour calculer le montant du plafonnement de l'ISF et de l'IFI."
    definition_period = YEAR

    def formula(foyer_fiscal, period):
        '''
        Voir le formulaire 2041-ISF-FCP
        https://www.impots.gouv.fr/portail/formulaire/2041-isf-fcp/fiche-de-calcul-du-plafonnement-isf
        Points à améliorer : impôts payés à l'étranger
        '''
        isf_ifi_avant_plaf = foyer_fiscal('isf_ifi_avant_plaf', period)
        irpp_economique = foyer_fiscal('irpp_economique', period)
        prelevement_forfaitaire_liberatoire = foyer_fiscal('prelevement_forfaitaire_liberatoire', period)
        prelevement_forfaitaire_unique_ir = foyer_fiscal('prelevement_forfaitaire_unique_ir', period)
        ir_pv_immo = foyer_fiscal('ir_pv_immo', period)
        crds_i = foyer_fiscal.members('crds', period)
        csg_i = foyer_fiscal.members('csg', period)
        crds = foyer_fiscal.sum(crds_i, role = FoyerFiscal.DECLARANT)
        csg = foyer_fiscal.sum(csg_i, role = FoyerFiscal.DECLARANT)
        prelevements_sociaux_revenus_capital_hors_csg_crds = foyer_fiscal('prelevements_sociaux_revenus_capital_hors_csg_crds', period)

        return (
            isf_ifi_avant_plaf
            - irpp_economique
            - prelevement_forfaitaire_liberatoire
            - prelevement_forfaitaire_unique_ir
            - ir_pv_immo
            - crds
            - csg
            - prelevements_sociaux_revenus_capital_hors_csg_crds
            )


class revenus_et_produits_plafonnement_isf_ifi(Variable):
    value_type = float
    entity = FoyerFiscal
    label = "Revenus et produits perçus, utilisés pour calculer le montant du plafonnement de l'ISF"
    definition_period = YEAR

    def formula(foyer_fiscal, period, parameters):
        '''
        Voir le formulaire 2041-ISF-FCP
        https://www.impots.gouv.fr/portail/formulaire/2041-isf-fcp/fiche-de-calcul-du-plafonnement-isf
        Points à améliorer : revenus de l'étranger et revenus exonérés
        '''
        salcho_imp_i = foyer_fiscal.members('revenu_assimile_salaire_apres_abattements', period)
        pen_net_i = foyer_fiscal.members('revenu_assimile_pension_apres_abattements', period)
        rente_viagere_titre_onereux_net = foyer_fiscal('rente_viagere_titre_onereux_net', period)
        revenu_categoriel_foncier = foyer_fiscal('revenu_categoriel_foncier', period)
        ric_i = foyer_fiscal.members('ric', period)
        rag_i = foyer_fiscal.members('rag', period)
        rpns_exon_i = foyer_fiscal.members('rpns_exon', period)
        rpns_pvct_i = foyer_fiscal.members('rpns_pvct', period)
        revenus_capitaux_prelevement_bareme = foyer_fiscal('revenus_capitaux_prelevement_bareme', period, options = [ADD])  # Supprimée à partir de 2018
        revenus_capitaux_prelevement_liberatoire = foyer_fiscal('revenus_capitaux_prelevement_liberatoire', period, options = [ADD])  # Supprimée à partir de 2018
        revenus_capitaux_prelevement_forfaitaire_unique_ir = foyer_fiscal('revenus_capitaux_prelevement_forfaitaire_unique_ir', period, options = [ADD])  # Existe à partir de 2018
        plus_values_base_large = foyer_fiscal('plus_values_base_large', period)
        assurance_vie_ps_exoneree_irpp_pl = foyer_fiscal('assurance_vie_ps_exoneree_irpp_pl', period)
        interets_pel_moins_12_ans_cel_i = foyer_fiscal.members('interets_pel_moins_12_ans_cel', period)
        livret_a_i = foyer_fiscal.members('livret_a', period.last_month)
        taux_livret_a = parameters(period).taxation_capital.epargne.livret_a.taux
        interets_livret_a_i = livret_a_i * taux_livret_a

        revenu_assimile_pension_apres_abattements = foyer_fiscal.sum(pen_net_i)
        rag = foyer_fiscal.sum(rag_i)
        ric = foyer_fiscal.sum(ric_i)
        rpns_exon = foyer_fiscal.sum(rpns_exon_i)
        rpns_pvct = foyer_fiscal.sum(rpns_pvct_i)
        revenu_assimile_salaire_apres_abattements = foyer_fiscal.sum(salcho_imp_i)
        interets_pel_moins_12_ans_cel = foyer_fiscal.sum(interets_pel_moins_12_ans_cel_i)
        interets_livret_a = foyer_fiscal.sum(interets_livret_a_i)

        montant = max_(
            0,
            revenu_assimile_salaire_apres_abattements
            + revenu_assimile_pension_apres_abattements
            + rente_viagere_titre_onereux_net
            + revenus_capitaux_prelevement_bareme
            + revenus_capitaux_prelevement_liberatoire
            + revenus_capitaux_prelevement_forfaitaire_unique_ir
            + ric
            + rag
            + rpns_exon
            + rpns_pvct
            + revenu_categoriel_foncier
            + plus_values_base_large
            + assurance_vie_ps_exoneree_irpp_pl
            + interets_pel_moins_12_ans_cel
            + interets_livret_a
            )

        return montant


class decote_isf_ifi(Variable):
    value_type = float
    entity = FoyerFiscal
    label = "Décote de l'ISF-IFI"
    definition_period = YEAR

    def formula_2018(foyer_fiscal, period, parameters):
        assiette_isf_ifi = foyer_fiscal('assiette_isf_ifi', period)
        decote = parameters(period).taxation_capital.impot_fortune_immobiliere_ifi_partir_2018.decote

        elig = (assiette_isf_ifi >= decote.borne_inferieure_decote) & (assiette_isf_ifi <= decote.borne_superieure_decote)
        lb = decote.parametre_calcul_decote - decote.taux_decote * assiette_isf_ifi
        return lb * elig

    def formula_2013(foyer_fiscal, period, parameters):
        assiette_isf_ifi = foyer_fiscal('assiette_isf_ifi', period)
        decote = parameters(period).taxation_capital.impot_solidarite_fortune_isf_1989_2017.decote

        elig = (assiette_isf_ifi >= decote.borne_inferieure_decote) & (assiette_isf_ifi <= decote.borne_superieure_decote)
        lb = decote.parametre_calcul_decote - decote.taux_decote * assiette_isf_ifi
        return lb * elig


class isf_ifi_apres_plaf(Variable):
    value_type = float
    entity = FoyerFiscal
    label = 'ISF-IFI après plafonnement'
    definition_period = YEAR

    def formula_2002_01_01(foyer_fiscal, period, parameters):
        total_impots_plafonnement_isf_ifi = foyer_fiscal('total_impots_plafonnement_isf_ifi', period)
        revenus_et_produits_plafonnement_isf_ifi = foyer_fiscal('revenus_et_produits_plafonnement_isf_ifi', period)
        isf_ifi_avant_plaf = foyer_fiscal('isf_ifi_avant_plaf', period)
        plaf = parameters(period).taxation_capital.impot_solidarite_fortune_isf_1989_2017.plaf

        # si ISF avant plafonnement n'excède pas seuil 1= la limitation du plafonnement ne joue pas
        # si entre les deux seuils; l'allègement est limité au 1er seuil
        # si ISF avant plafonnement est supérieur au 2nd seuil, l'allègement qui résulte du plafonnement
        # est limité à 50% de l'ISF
        plafonnement = max_(total_impots_plafonnement_isf_ifi - revenus_et_produits_plafonnement_isf_ifi, 0)
        limitationplaf = (
            (isf_ifi_avant_plaf <= plaf.plaf.seuil1) * plafonnement
            + (plaf.plaf.seuil1 <= isf_ifi_avant_plaf) * (isf_ifi_avant_plaf <= plaf.plaf.seuil2) * min_(plafonnement, plaf.plaf.seuil1)
            + (isf_ifi_avant_plaf >= plaf.plaf.seuil2) * min_(isf_ifi_avant_plaf * plaf.plafonnement_plafonnement, plafonnement)
            )
        return max_(isf_ifi_avant_plaf - limitationplaf, 0)

    def formula_2012_01_01(foyer_fiscal, period):
        '''
        Plafonnement supprimé pour l'année 2012
        '''
        isf_ifi_avant_plaf = foyer_fiscal('isf_ifi_avant_plaf', period)
        return isf_ifi_avant_plaf

    def formula_2013_01_01(foyer_fiscal, period, parameters):
        total_impots_plafonnement_isf_ifi = foyer_fiscal('total_impots_plafonnement_isf_ifi', period)
        revenus_et_produits_plafonnement_isf_ifi = foyer_fiscal('revenus_et_produits_plafonnement_isf_ifi', period)
        isf_ifi_avant_plaf = foyer_fiscal('isf_ifi_avant_plaf', period)
        plaf = parameters(period).taxation_capital.impot_solidarite_fortune_isf_1989_2017.plaf

        plafond = max_(0, total_impots_plafonnement_isf_ifi - plaf.plafonnement_taux_imposition * revenus_et_produits_plafonnement_isf_ifi)  # case 9PV sur le formulaire 2042C des revenus 2013 aux revenus 2016
        return max_(isf_ifi_avant_plaf - plafond, 0)

    def formula_2018_01_01(foyer_fiscal, period, parameters):
        total_impots_plafonnement_isf_ifi = foyer_fiscal('total_impots_plafonnement_isf_ifi', period)
        revenus_et_produits_plafonnement_isf_ifi = foyer_fiscal('revenus_et_produits_plafonnement_isf_ifi', period)
        isf_ifi_avant_plaf = foyer_fiscal('isf_ifi_avant_plaf', period)
        plaf = parameters(period).taxation_capital.impot_fortune_immobiliere_ifi_partir_2018.plaf

        plafond = max_(0, total_impots_plafonnement_isf_ifi - plaf.plafonnement_taux_imposition * revenus_et_produits_plafonnement_isf_ifi)  # case 9PV sur le formulaire 2042C des revenus 2013 aux revenus 2016
        return max_(isf_ifi_avant_plaf - plafond, 0)


class isf_ifi(Variable):
    value_type = float
    entity = FoyerFiscal
    label = "Montant d'ISF-IFI"
    reference = 'https://www.legifrance.gouv.fr/codes/id/LEGISCTA000036385039/'
    definition_period = YEAR

    def formula(foyer_fiscal, period):
        b4rs = foyer_fiscal('b4rs', period)
        isf_ifi_apres_plaf = foyer_fiscal('isf_ifi_apres_plaf', period)
        return min_(-(isf_ifi_apres_plaf - b4rs), 0)


# BOUCLIER FISCAL

# calcul de l'ensemble des revenus du contribuable


# TODO: à reintégrer dans irpp
class rvcm_plus_abat(Variable):
    value_type = float
    entity = FoyerFiscal
    label = 'rvcm_plus_abat'
    definition_period = YEAR
    end = '2010-12-31'

    def formula(foyer_fiscal, period, parameters):
        '''
        Revenu catégoriel avec abattement de 40% réintégré.
        '''
        rev_cat_rvcm = foyer_fiscal('revenu_categoriel_capital', period)
        rfr_rvcm_abattements_a_reintegrer = foyer_fiscal('rfr_rvcm_abattements_a_reintegrer', period)

        return rev_cat_rvcm + rfr_rvcm_abattements_a_reintegrer


class maj_cga(Variable):
    value_type = float
    entity = Individu
    label = 'Majoration pour non adhésion à un centre de gestion agréé (pour chaque individu du foyer)'
    definition_period = YEAR

    # TODO: à reintégrer dans irpp (et vérifier au passage que frag_impo est dans la majo_cga
    def formula(individu, period, parameters):
        frag_impo = individu('frag_impo', period)
        nrag_impg = individu('nrag_impg', period)
        nbic_impn = individu('nbic_impn', period)
        nbic_imps = individu('nbic_imps', period)
        nbic_defn = individu('nbic_defn', period)
        nbic_defs = individu('nbic_defs', period)
        nacc_impn = individu('nacc_impn', period)
        nacc_meup = individu('nacc_meup', period)
        nacc_defn = individu('nacc_defn', period)
        nacc_defs = individu('nacc_defs', period)
        nbnc_impo = individu('nbnc_impo', period)
        nbnc_defi = individu('nbnc_defi', period)
        rpns = parameters(period).impot_revenu.calcul_revenus_imposables.rpns

        nbic_timp = (nbic_impn + nbic_imps) - (nbic_defn + nbic_defs)

        # C revenus industriels et commerciaux non professionnels
        # (revenus accesoires du foyers en nomenclature INSEE)
        nacc_timp = max_(0, (nacc_impn + nacc_meup) - (nacc_defn + nacc_defs))

        # régime de la déclaration contrôlée ne bénéficiant pas de l'abattement association agréée
        nbnc_timp = nbnc_impo - nbnc_defi

        # Totaux
        ntimp = nrag_impg + nbic_timp + nacc_timp + nbnc_timp

        return max_(0, rpns.cga_taux2 * (ntimp + frag_impo))


class bouclier_rev(Variable):
    value_type = float
    entity = FoyerFiscal
    label = 'bouclier_rev'
    end = '2010-12-31'
    definition_period = YEAR

    def formula_2006(foyer_fiscal, period, parameters):
        '''
        Total des revenus sur l'année 'n' net de charges
        '''
        rbg = foyer_fiscal('rbg', period)
        csg_patrimoine_deductible_ir = foyer_fiscal('csg_patrimoine_deductible_ir', period)
        rvcm_plus_abat = foyer_fiscal('rvcm_plus_abat', period)
        revenus_capitaux_prelevement_liberatoire = foyer_fiscal('revenus_capitaux_prelevement_liberatoire', period)
        rev_exo = foyer_fiscal('rev_exo', period)  # noqa F841
        rev_or = foyer_fiscal('rev_or', period)
        pensions_alimentaires_deduites = foyer_fiscal('pensions_alimentaires_deduites', period)
        cd_eparet = foyer_fiscal('cd_eparet', period)

        maj_cga_i = foyer_fiscal.members('maj_cga', period)  # noqa F841
        maj_cga = foyer_fiscal.sum(maj_cga)  # noqa F841

        # TODO: réintégrer les déficits antérieur
        # TODO: intégrer les revenus soumis au prélèvement libératoire
        # deficit_ante =

        # # Revenus
        # TODO: UNUSED ?
        frac_rvcm_rfr = 0.7 * rvcm_plus_abat  # noqa F841
        # # revenus distribués?
        # # A majorer de l'abatt de 40% - montant brut en cas de PFL
        # # pour le calcul de droit à restitution : prendre 0.7*montant_brut_rev_dist_soumis_au_barème
        # rev_bar = rbg - maj_cga - csg_patrimoine_deductible_ir - deficit_ante
        rev_bar = rbg - maj_cga - csg_patrimoine_deductible_ir

    # # TODO: AJOUTER : indemnités de fonction percus par les élus- revenus soumis à régimes spéciaux

        # Revenu soumis à l'impôt sur le revenu forfaitaire
        rev_lib = revenus_capitaux_prelevement_liberatoire
        # TODO: AJOUTER plus-values immo et moins values?

        # Revenus exonérés d'IR réalisés en France et à l'étranger
        # rev_exo = primes_pel + primes_cel + rente_pea + int_livrets + plus_values_per

        # TODO: proposer à l'utilisateur des taux de réference- PER, PEA, PEL,...
        # sommes investis- calculer les plus_values annuelles et prendre en compte pour rev_exo?
        # revenus soumis à la taxe forfaitaire sur les métaux précieux : rev_or

        # revenus = rev_bar + rev_lib + rev_exo + rev_or
        revenus = rev_bar + rev_lib + rev_or

        # # CHARGES
        # Pension alimentaires
        # Cotisations ou primes versées au titre de l'épargne retraite

        charges = pensions_alimentaires_deduites + cd_eparet

        return revenus - charges


class bouclier_imp_gen(Variable):  # # ajouter CSG- CRDS
    value_type = float
    entity = FoyerFiscal
    label = 'bouclier_imp_gen'
    end = '2010-12-31'
    definition_period = YEAR

    def formula_2006(foyer_fiscal, period, parameters):
        irpp = foyer_fiscal('irpp', period)
        tax_fonc = foyer_fiscal('tax_fonc', period)
        isf_ifi = foyer_fiscal('isf_ifi', period)
        csg_deductible_salaire_i = foyer_fiscal.members('csg_deductible_salaire', period)
        csg_imposable_salaire_i = foyer_fiscal.members('csg_imposable_salaire', period)
        crds_salaire_i = foyer_fiscal.members('crds_salaire', period)
        csg_imposable_chomage_i = foyer_fiscal.members('csg_imposable_chomage', period)
        csg_deductible_chomage_i = foyer_fiscal.members('csg_deductible_chomage', period)
        csg_deductible_retraite_i = foyer_fiscal.members('csg_deductible_retraite', period)
        csg_imposable_retraite_i = foyer_fiscal.members('csg_imposable_retraite', period)
        prelevement_forfaitaire_liberatoire = foyer_fiscal('prelevement_forfaitaire_liberatoire', period)

        prelevements_sociaux_revenus_capital = foyer_fiscal('prelevements_sociaux_revenus_capital', period)
        crds_salaire = foyer_fiscal.sum(crds_salaire_i)
        csg_deductible_chomage = foyer_fiscal.sum(csg_deductible_chomage_i)
        csg_imposable_chomage = foyer_fiscal.sum(csg_imposable_chomage_i)
        csg_deductible_salaire = foyer_fiscal.sum(csg_deductible_salaire_i)
        csg_imposable_salaire = foyer_fiscal.sum(csg_imposable_salaire_i)
        csg_deductible_retraite = foyer_fiscal.sum(csg_deductible_retraite_i)
        csg_imposable_retraite = foyer_fiscal.sum(csg_imposable_retraite_i)

        taxe_habitation_i = foyer_fiscal.members.menage('taxe_habitation', period)  # noqa F841
        taxe_habitation = foyer_fiscal.sum(taxe_habitation_i, role = Menage.PERSONNE_DE_REFERENCE)

        # # ajouter Prelèvements sources/ libé
        # # ajouter crds rstd
        # # impôt sur les plus-values immo et cession de fonds de commerce
        imp1 = prelevements_sociaux_revenus_capital + csg_deductible_salaire + csg_deductible_chomage + crds_salaire + csg_deductible_retraite + prelevement_forfaitaire_liberatoire
        '''
        Impôts payés en l'année 'n' au titre des revenus réalisés sur l'année 'n'
        '''
        imp2 = irpp + isf_ifi + taxe_habitation + tax_fonc + csg_imposable_salaire + csg_imposable_chomage + csg_imposable_retraite
        '''
        Impôts payés en l'année 'n' au titre des revenus réalisés en 'n-1'
        '''
        return imp1 + imp2


class restitutions(Variable):
    value_type = float
    entity = FoyerFiscal
    label = 'restitutions'
    end = '2010-12-31'
    definition_period = YEAR

    def formula_2006(foyer_fiscal, period, parameters):
        '''
        Restitutions d'impôt sur le revenu et degrèvements percus en l'année 'n'
        '''
        ppe = foyer_fiscal('ppe', period)
        restit_imp = foyer_fiscal('restit_imp', period)

        return ppe + restit_imp


class bouclier_sumimp(Variable):
    value_type = float
    entity = FoyerFiscal
    label = 'bouclier_sumimp'
    end = '2010-12-31'
    definition_period = YEAR

    def formula_2006(foyer_fiscal, period, parameters):
        '''
        Somme totale des impôts moins restitutions et degrèvements
        '''
        bouclier_imp_gen = foyer_fiscal('bouclier_imp_gen', period)
        restitutions = foyer_fiscal('restitutions', period)

        return -bouclier_imp_gen + restitutions


class bouclier_fiscal(Variable):
    value_type = float
    entity = FoyerFiscal
    label = 'bouclier_fiscal'
    end = '2010-12-31'
    reference = 'http://fr.wikipedia.org/wiki/Bouclier_fiscal'
    definition_period = YEAR

    def formula_2006(foyer_fiscal, period, parameters):
        bouclier_sumimp = foyer_fiscal('bouclier_sumimp', period)
        bouclier_rev = foyer_fiscal('bouclier_rev', period)
        bouclier_fiscal = parameters(period).impot_revenu.calcul_impot_revenu.bouclier_fiscal

        return max_(0, bouclier_sumimp - (bouclier_rev * bouclier_fiscal.taux))



from openfisca_france.model.base import *


class professionnalisation(Variable):
    value_type = bool
    entity = Individu
    label = "L'individu est en contrat de professionnalisation"
    reference = 'https://www.legifrance.gouv.fr/codes/id/LEGIARTI000031088014/2016-01-01/'
    definition_period = MONTH
    set_input = set_input_dispatch_by_period

    def formula(individu, period, parameters):
        age = individu('age', period)
        ass = individu('ass', period)
        rsa = individu.famille('rsa', period)
        aah = individu('aah', period)

        age_condition = (16 <= age) * (age < 25)
        dummy_ass = ass > 0
        dummy_rmi = rsa > 0
        dummy_aah = aah > 0

        return (age_condition + dummy_ass + dummy_aah + dummy_rmi) > 0


class qualifie(Variable):
    value_type = bool
    entity = Individu
    label = 'Etat du niveau de formation ou de qualification avant le contrat de professionnalisation'
    definition_period = MONTH
    set_input = set_input_dispatch_by_period
    reference = 'https://travail-emploi.gouv.fr/formation-professionnelle/formation-en-alternance-10751/contrat-de-professionnalisation'


class remuneration_professionnalisation(Variable):
    value_type = float
    entity = Individu
    label = "Rémunération de l'apprenti sous contrat de professionalisation"
    reference = 'http://www.apce.com/pid927/contrat-d-apprentissage.html?espace=1&tp=1&pagination=2'
    definition_period = MONTH
    set_input = set_input_divide_by_period

    #  La rémunération minimale varie en fonction de l'âge et du niveau de qualification des bénéficiaires des contrats
    #  de professionnalisation :
    #
    #  Pour les personnes de moins de 21 ans :
    #  au minimum 55 % du Smic,
    #  au minimum 65 % du Smic si le jeune est titulaire d'une qualification au moins égale au baccalauréat
    #  professionnel ou d'un titre ou d'un diplôme à finalité professionnelle de même niveau.
    #
    #  Pour les personnes ayant entre 21 et 25 ans :
    #  au minimum 70 % du Smic,
    #  au minimum 80 % du Smic si le bénéficiaire est titulaire d'une qualification au moins égale à celle d'un
    #  baccalauréat professionnel ou d'un titre/diplôme à finalité professionnelle de même niveau.
    #
    #  Pour les personnes âgées de plus de 26 ans :
    #  au minimum le Smic,
    #  au minimum 85 % du salaire minimum prévu par la convention ou l'accord de branche auquel est soumise
    #  l'entreprise.

    def formula(individu, period, parameters):
        age = individu('age', period)
        smic = parameters(period).marche_travail.salaire_minimum.smic.smic_b_horaire * 52 * 35 / 12
        professionnalisation = individu('professionnalisation', period)
        qualifie = individu('qualifie', period)
        salaire_en_smic = [
            dict(
                part_de_smic_by_qualification = {
                    'non_qualifie': .55,
                    'qualifie': .65
                    },
                age_min = 16,
                age_max = 21,
                ),
            dict(
                part_de_smic_by_qualification = {
                    1: .70,
                    },
                age_min = 21,
                age_max = 25,
                ),
            dict(
                part_de_smic_by_qualification = {
                    1: 1.0,
                    },
                age_min = 26,
                age_max = 99
                )
            ]

        taux_smic = age * 0.0
        for age_interval in salaire_en_smic:
            age_condition = (age_interval['age_min'] <= age) * (age <= age_interval['age_max'])
            taux_smic[age_condition] = sum([
                (qualifie[age_condition] == qualification) * part_de_smic
                for qualification, part_de_smic in age_interval['part_de_smic_by_qualification'].items()
                ])
        return taux_smic * smic * professionnalisation


class exoneration_cotisations_employeur_professionnalisation(Variable):
    value_type = float
    entity = Individu
    label = "Exonération de cotisations patronales pour l'emploi d'un apprenti"
    reference = 'http://www.apce.com/pid927/contrat-d-apprentissage.html?espace=1&tp=1&pagination=2'
    definition_period = MONTH
    set_input = set_input_divide_by_period

    #  Exonération de cotisations sociales patronales d'assurance maladie-maternité, de vieillesse de base,
    #  d'invalidité-décès et d'allocations familiales au titre des rémunérations versées aux demandeurs d'emploi de
    #  plus de 45 ans
    #
    #  Les salariés en contrat de professionnalisation ne sont pas comptabilisés dans l'effectif de l'entreprise pendant
    #  la durée du contrat s'il est à durée déterminée ou pendant l'action de professionnalisation si le contrat est à
    #  durée indéterminée.
    #  Remboursement de certaines dépenses par les organismes collecteurs paritaires agréés (OPCA)
    #  Aide forfaitaire versée par Pôle emploi pour l'embauche d'un demandeur d'emploi de 26 ans et plus
    #  En cas d'embauche d'un demandeur d'emploi de 26 ans et plus, l'employeur peut bénéficier d'une aide forfaitaire
    #  (AFE) d'un montant maximum de 2 000 euros par bénéficiaire. Pour les salariés à temps partiel, le montant de
    #  l'aide est proratisé en fonction du temps de travail effectif.
    #  Aide spécifique de 686 euros par accompagnement et pour une année pleine est attribuée sous certaines conditions
    #  aux groupements d'employeurs qui organisent dans le cadre des contrats de professionnalisation

    def formula(individu, period, parameters):
        age = individu('age', period)
        mmid_employeur = individu('mmid_employeur', period)
        famille = individu('famille', period)
        vieillesse_plafonnee_employeur = individu('vieillesse_plafonnee_employeur', period)
        # FIXME: correspond bien à vieillesse de base ?
        cotisations_exonerees = mmid_employeur + famille + vieillesse_plafonnee_employeur

        return cotisations_exonerees * (age > 45)
        # FIXME: On est bien d'accord qu'il y a les exos uniquement pour les
        # plus de 45 ans?

# TODO: vérifier aucun avantage pour l'employé ??



from typing import Callable, Tuple

import numpy
import pytest

from ..cache import tax_benefit_system
from openfisca_core.simulations import Simulation
from openfisca_france.reforms.plfr2014 import plfr2014


@pytest.fixture
def simulations(new_scenario) -> Callable[..., Tuple[Simulation]]:
    def _simulations(year: int) -> Tuple[Simulation]:
        scenario = new_scenario(
            reform = plfr2014(tax_benefit_system),
            count = 2,
            _max = 18000,
            _min = 0,
            name = 'salaire_imposable',
            year = year,
            people = 1,
            )

        return scenario.new_simulation(use_baseline = True), scenario.new_simulation()

    return _simulations


def test_plf2014(simulations, year = 2013):
    actual_simulation, reform_simulation = simulations(year)
    actual_impots = actual_simulation.calculate('impots_directs', year)
    reform_impots = reform_simulation.calculate('impots_directs', year)
    assert numpy.all(numpy.equal(actual_impots, reform_impots))



import logging
from openfisca_france.model.base import *

log = logging.getLogger(__name__)


# Prélèvements sociaux sur les revenus du capital
# (dispositifs codés à partir de 2013 : cf. doctring de la variable assiette_csg_revenus_capital)


# 1. Définition de variables associées aux revenus du capital soumis aux prélèvements sociaux mais
#    ni au barème de l'impôt sur le revenu, ni au prélèvement forfaitaire libératoire
#    (et donc non présents sur les déclarations de revenu)


class interets_plan_epargne_logement_moins_de_12_ans_ouvert_avant_2018(Variable):
    '''
    NB :
    (1) Cette variable est définie indépendemment de epargne_revenus_non_imposables
    (2) Les intérêts des PEL de plus de 12 ans sont imposables à l'impôt sur le revenu, et donc déjà présents dans les cases 2TR ou 2FA (attention: formulaire IR 2019 sur revenus 2018 non connu au moment de coder cette variable)
    '''
    value_type = float
    entity = Individu
    label = 'Intérêts des plans épargne logement (PEL) de moins de 12 ans ouverts avant le 1er janvier 2018'
    definition_period = YEAR


class interets_plan_epargne_logement_moins_de_12_ans_ouvert_a_partir_de_2018(Variable):
    '''
    NB :
    (1) Cette variable est définie indépendemment de epargne_revenus_non_imposables
    (2) Les intérêts des PEL de plus de 12 ans sont imposables à l'impôt sur le revenu, et donc déjà présents dans les cases 2TR ou 2FA (attention: formulaire IR 2019 sur revenus 2018 non connu au moment de coder cette variable)
    '''
    value_type = float
    entity = Individu
    label = 'Intérêts des plans épargne logement (PEL) de moins de 12 ans ouverts à partir du 1er janvier 2018'
    definition_period = YEAR


class interets_compte_epargne_logement_ouvert_avant_2018(Variable):
    ''' NB : Cette variable est définie indépendemment de epargne_revenus_non_imposables '''
    value_type = float
    entity = Individu
    label = 'Intérêts des comptes épargne logement (CEL) ouverts avant le 1er janvier 2018'
    definition_period = YEAR


class interets_compte_epargne_logement_ouvert_a_partir_de_2018(Variable):
    ''' NB : Cette variable est définie indépendemment de epargne_revenus_non_imposables '''
    value_type = float
    entity = Individu
    label = 'Intérêts des comptes épargne logement (CEL) ouverts à partir du 1er janvier 2018'
    definition_period = YEAR


class interets_pel_moins_12_ans_cel(Variable):
    ''' NB : Cette variable est définie indépendemment de epargne_revenus_non_imposables '''
    value_type = float
    entity = Individu
    label = 'Intérêts des plans épargne logement (PEL) de moins de 12 ans et des comptes épargne logement (CEL)'
    definition_period = YEAR

    def formula(individu, period):

        interets_plan_epargne_logement_moins_de_12_ans_ouvert_avant_2018 = individu('interets_plan_epargne_logement_moins_de_12_ans_ouvert_avant_2018', period)
        interets_compte_epargne_logement_ouvert_avant_2018 = individu('interets_compte_epargne_logement_ouvert_avant_2018', period)

        return (
            interets_plan_epargne_logement_moins_de_12_ans_ouvert_avant_2018
            + interets_compte_epargne_logement_ouvert_avant_2018
            )

    def formula_2018_01_01(individu, period):

        interets_plan_epargne_logement_moins_de_12_ans_ouvert_avant_2018 = individu('interets_plan_epargne_logement_moins_de_12_ans_ouvert_avant_2018', period)
        interets_compte_epargne_logement_ouvert_avant_2018 = individu('interets_compte_epargne_logement_ouvert_avant_2018', period)
        interets_plan_epargne_logement_moins_de_12_ans_ouvert_a_partir_de_2018 = individu('interets_plan_epargne_logement_moins_de_12_ans_ouvert_a_partir_de_2018', period)
        interets_compte_epargne_logement_ouvert_a_partir_de_2018 = individu('interets_compte_epargne_logement_ouvert_a_partir_de_2018', period)

        return (
            interets_plan_epargne_logement_moins_de_12_ans_ouvert_avant_2018
            + interets_compte_epargne_logement_ouvert_avant_2018
            + interets_plan_epargne_logement_moins_de_12_ans_ouvert_a_partir_de_2018
            + interets_compte_epargne_logement_ouvert_a_partir_de_2018
            )


class assurance_vie_ps_exoneree_irpp_pl(Variable):
    value_type = float
    entity = FoyerFiscal
    label = "Produits d'assurance-vie exonérés d'impôt sur le revenu et de prélèvement libératoire mais soumis aux prélèvements sociaux"
    definition_period = YEAR


# 2. Assiette des revenus du capital soumis à la CSG (valable pour les autres prélèvements sociaux)

class assiette_csg_plus_values(Variable):
    value_type = float
    entity = FoyerFiscal
    label = 'Assiette des plus-values soumis à la CSG'
    definition_period = YEAR

    def formula(foyer_fiscal, period):
        '''
        Attention : cette formule est susceptible de contenir des erreurs pour les années avant 2013 (cf. commentaires sur assiette_csg_revenus_capital)

        Notes sur le champ de cette variable :
            Cette assiette de plus-values est partielle. De nombreux types de plus-values sont
            manquants, rien que parmi les plus-values déclarées dans la déclaration de revenus
            au titre de l'impôt sur le revenu. Ceci s'explique par la complexité de la législation sur
            la prise en compte des plus-values dans le calcul des prélèvements sociaux (exemple : plus-values
            réalisées sur PEA taxables selon les règles en vigueur l'année de réalisation et non l'année de retrait,
            plus-values taxées après report, et taxables parfois selon la législation en vigueur au moment du report et non
            pas au moment de la taxation, etc.).
            Critère de choix : on part des dénombrements fiscaux de la déclaration 2042 des revenus 2016
            (montants totaux déclarés au niveau national pour chaque case), et on constate que les cases 3VH, 3VG,
            3SG, 3SL, 3VA, 3VB, 3VO, 3VP, 3VZ, 3VW, 3WG, 3WH et 3WM représentent à elles seules 88% des
            montants associés aux plus-values (cases de classe 3). On se limite donc aux
            plus-values associées à ces cases. Parmi ces cases, certaines ne donnent pas lieu à une imposition
            pendant l'année courante du fait d'un report ou sursis d'imposition (3WG, 3WH, 3WM), ou ne sont
            tout simplement pas comptabilisées dans l'assiette CSG (3VH, 3VW). Au total, le nombre de cases
            est fortement réduit, mais en ne perdant potentiellement qu'une faible partie des plus-values présentes
            dans l'assiette CSG (même en excluant ces cases importantes mais non sujettes à imposition, les cases
            restantes représentent 82% des montants de l'ensemble des cases de classe 3 (hors cases supprimées).
            NB : en plus des cases les plus importantes, on inclut aussi 3WE, car les abattements
            renseignés dans les cases 3SG et 3SL peuvent aussi être associés à des montants nets renseignés en 3WE.

        Notes concernant les plus-values immobilières :
            (1) Les plus-values immobilières déclarées en 3VZ sont les plus-values nettes sousmises à
                l'impôt sur le revenu. Or, les prélèvements sociaux sont appliqués aussi sur une valeure
                nette, mais déterminée via des abattements différents de ceux de l'impôt sur le revenu
                (cf. formulaire 2048-IMM de 2018 par exemple pour une explication). On ignore ces
                différences d'abattement, et on suppose que la valeur nette au sens des prélèvements
                sociaux est la même que celle au sens de l'impôt sur le revenu.
            (2) On ne compte pas la case 3VW dans la base soumise aux prélèvements sociaux. Ce montant,
                exonéré de l'impôt sur le revenu, semble être exonéré aussi des prélèvements sociaux,
                même s'il est déclaré dans la déclaration de revenus : cf. art. L136-7 du CSS, qui
                cite l'art. 150 U du CGI.
        '''

        # Plus-values mobilières brutes (avant abattement)
        f3vg = foyer_fiscal('f3vg', period)
        f3sg = foyer_fiscal('f3sg', period)
        f3sl = foyer_fiscal('f3sl', period)
        f3va_2014 = foyer_fiscal('f3va_2014', period)
        f3we = foyer_fiscal('f3we', period)
        f3vt = foyer_fiscal('f3vt', period)
        rpns_pvce_i = foyer_fiscal.members('rpns_pvce', period)
        rpns_pvce = foyer_fiscal.sum(rpns_pvce_i)

        # Plus-values immobilières
        f3vz = foyer_fiscal('f3vz', period)

        return f3vg + f3sg + f3sl + f3va_2014 + f3vz + f3we + f3vt + rpns_pvce

    def formula_2015_01_01(foyer_fiscal, period, parameters):
        '''
        Notes concernant les plus-values immobilières : cf. formule commençant en 2013
        '''

        # Plus-values mobilières brutes (avant abattement)
        f3vg = foyer_fiscal('f3vg', period)
        f3sg = foyer_fiscal('f3sg', period)
        f3sl = foyer_fiscal('f3sl', period)
        f3va_2016_i = foyer_fiscal.members('f3va_2016', period)
        f3va_2016 = foyer_fiscal.sum(f3va_2016_i)
        f3we = foyer_fiscal('f3we', period)
        f3vt = foyer_fiscal('f3vt', period)
        rpns_pvce_i = foyer_fiscal.members('rpns_pvce', period)
        rpns_pvce = foyer_fiscal.sum(rpns_pvce_i)

        # Plus-values immobilières
        f3vz = foyer_fiscal('f3vz', period)

        return f3vg + f3sg + f3sl + f3va_2016 + f3vz + f3we + f3vt + rpns_pvce

    def formula_2017_01_01(foyer_fiscal, period, parameters):
        '''
        Notes concernant les plus-values immobilières : cf. formule commençant en 2013
        '''

        # Plus-values mobilières brutes (avant abattement)
        f3vg = foyer_fiscal('f3vg', period)
        f3sg = foyer_fiscal('f3sg', period)
        f3sl = foyer_fiscal('f3sl', period)
        f3va = foyer_fiscal('f3va', period)
        f3we = foyer_fiscal('f3we', period)
        f3ua = foyer_fiscal('f3ua', period)
        f3vt = foyer_fiscal('f3vt', period)
        f3pi = foyer_fiscal('f3pi', period)
        rpns_pvce_i = foyer_fiscal.members('rpns_pvce', period)
        rpns_pvce = foyer_fiscal.sum(rpns_pvce_i)

        # Plus-values immobilières
        f3vz = foyer_fiscal('f3vz', period)

        return f3vg + f3sg + f3sl + f3va + f3ua + f3vz + f3we + f3vt + f3pi + rpns_pvce

    def formula_2018_01_01(foyer_fiscal, period, parameters):

        # Plus-values mobilières brutes (avant abattement)
        f3vg = foyer_fiscal('f3vg', period)
        f3we = foyer_fiscal('f3we', period)
        f3ua = foyer_fiscal('f3ua', period)
        f3vd_i = foyer_fiscal.members('f3vd', period)
        f3vi_i = foyer_fiscal.members('f3vi', period)
        f3vf_i = foyer_fiscal.members('f3vf', period)
        f3sj = foyer_fiscal('f3sj', period)
        f3tj = foyer_fiscal('f3tj', period)
        f3sk = foyer_fiscal('f3sk', period)
        f3vm = foyer_fiscal('f3vm', period)
        f3vt = foyer_fiscal('f3vt', period)
        f3wi = foyer_fiscal('f3wi', period)
        f3wj = foyer_fiscal('f3wj', period)
        rpns_pvce_i = foyer_fiscal.members('rpns_pvce', period)
        f3pi = foyer_fiscal('f3pi', period)

        rpns_pvce = foyer_fiscal.sum(rpns_pvce_i)
        f3vd = foyer_fiscal.sum(f3vd_i)
        f3vi = foyer_fiscal.sum(f3vi_i)
        f3vf = foyer_fiscal.sum(f3vf_i)

        # Plus-values immobilières
        f3vz = foyer_fiscal('f3vz', period)

        return f3vg + f3ua + f3vz + f3we + rpns_pvce + f3sj + f3sk + f3vm + f3vt + f3wi + f3wj + f3pi + f3tj + f3vd + f3vi + f3vf

    def formula_2019_01_01(foyer_fiscal, period, parameters):

        # Plus-values mobilières brutes (avant abattement)
        f3vg = foyer_fiscal('f3vg', period)
        f3we = foyer_fiscal('f3we', period)
        f3ua = foyer_fiscal('f3ua', period)
        f3vd_i = foyer_fiscal.members('f3vd', period)
        f3vi_i = foyer_fiscal.members('f3vi', period)
        f3vf_i = foyer_fiscal.members('f3vf', period)
        f3sj = foyer_fiscal('f3sj', period)
        f3tj = foyer_fiscal('f3tj', period)
        f3sk = foyer_fiscal('f3sk', period)
        f3vm = foyer_fiscal('f3vm', period)
        f3vt = foyer_fiscal('f3vt', period)
        f3wi = foyer_fiscal('f3wi', period)
        f3wj = foyer_fiscal('f3wj', period)
        rpns_pvce_i = foyer_fiscal.members('rpns_pvce', period)
        f3pi = foyer_fiscal('f3pi', period)
        f3an = foyer_fiscal('f3an', period)

        rpns_pvce = foyer_fiscal.sum(rpns_pvce_i)
        f3vd = foyer_fiscal.sum(f3vd_i)
        f3vi = foyer_fiscal.sum(f3vi_i)
        f3vf = foyer_fiscal.sum(f3vf_i)

        # Plus-values immobilières
        f3vz = foyer_fiscal('f3vz', period)

        return f3vg + f3ua + f3vz + f3we + rpns_pvce + f3sj + f3sk + f3vm + f3vt + f3wi + f3wj + f3pi + f3tj + f3an + f3vd + f3vi + f3vf


class assiette_csg_revenus_capital(Variable):
    value_type = float
    entity = FoyerFiscal
    label = 'Assiette des revenus du capital soumis à la CSG'
    definition_period = YEAR

    def formula(foyer_fiscal, period, parameters):
        '''
        Hypothèses dérrière ce calcul :
            (1) On ne distingue pas la CSG sur les revenus du patrimoine (art. L136-6 du CSS)
                de celle sur les revenus de placement (art. L136-6 du CSS)
                ATTENTION : Les taux de la CSG et de l'ensemble des prélèvements sociaux sont identiques pour
                ces deux types de revenu depuis 2013 seulement, la formule devrait donc être corrigée pour les années avant 2013.
            (2) Le timing de la soumission des intérêts des PEL et CEL aux prélèvements sociaux
                est complexe. Cette soumission peut se faire annuellement, ou en cumulé, et ce
                en fonction de différents paramètres. Mais on ne prend pas en compte cette fonctionnalité.
        NB : catégorie(s) de revenu non encore incluse(s) dans cette assiette : épargne salariale
        Note à partir de 2018 : du fait du PFU, la base des revenus du capital au titre de l'impôt sur le revenu se rapproche de la base au titre des prélèvements sociaux,
                                d'où le fait qu'on utilise cette variable. En revanche, concernant les prêts participatifs, le montant au titre de l'impôt sur le revenu
                                forfaitaire est le montant net des pertes, alors que celui soumis au titre des prélèvements sociaux est le montant brut. Cependant,
                                la case 2TT contient le montant des intérêts de ces prêts après déduction de ces pertes. Donc, on est contraint de prendre un montant net.
        '''

        # Revenus du capital présents dans la section 2 de la déclaration de revenus
        revenus_capitaux_prelevement_bareme = foyer_fiscal('revenus_capitaux_prelevement_bareme', period, options = [ADD])
        revenus_capitaux_prelevement_liberatoire = foyer_fiscal('revenus_capitaux_prelevement_liberatoire', period, options = [ADD])
        revenus_capitaux_prelevement_forfaitaire_unique_ir = foyer_fiscal('revenus_capitaux_prelevement_forfaitaire_unique_ir', period, options = [ADD])

        # Rentes viagères à titre onéreux
        rente_viagere_titre_onereux_net = foyer_fiscal('rente_viagere_titre_onereux_net', period)

        # Revenus des produits d'épargne logement
        interets_pel_moins_12_ans_cel_i = foyer_fiscal.members('interets_pel_moins_12_ans_cel', period)
        interets_pel_moins_12_ans_cel = foyer_fiscal.sum(interets_pel_moins_12_ans_cel_i)

        # Revenus fonciers
        rev_cat_rfon = foyer_fiscal('revenu_categoriel_foncier', period)

        # Plus-values
        assiette_csg_plus_values = foyer_fiscal('assiette_csg_plus_values', period)

        # produits d'assurance-vie exonérés d'impôt sur le revenu et de prélèvement forfaitaire libératoire (et donc non présents dans revenus_capitaux_prelevement_bareme et revenus_capitaux_prelevement_liberatoire)
        assurance_vie_ps_exoneree_irpp_pl = foyer_fiscal('assurance_vie_ps_exoneree_irpp_pl', period)

        # Crédits d'impôt sur valeurs étrangères déduits de la base CSG
        credits_impot_sur_valeurs_etrangeres = foyer_fiscal('credits_impot_sur_valeurs_etrangeres', period)

        return max_(
            revenus_capitaux_prelevement_bareme
            + revenus_capitaux_prelevement_liberatoire
            + revenus_capitaux_prelevement_forfaitaire_unique_ir
            + rente_viagere_titre_onereux_net
            + interets_pel_moins_12_ans_cel
            + rev_cat_rfon
            + assiette_csg_plus_values
            + assurance_vie_ps_exoneree_irpp_pl
            - credits_impot_sur_valeurs_etrangeres,
            0
            )


# 3. Variables de prélèvements sociaux sur les revenus du capital

class csg_revenus_capital(Variable):
    value_type = float
    entity = FoyerFiscal
    label = 'CSG sur les revenus du capital'
    definition_period = YEAR

    def formula(foyer_fiscal, period, parameters):
        '''
        Attention : Pour les années avant 2013, cette formule n'est pas entièrement correcte car le taux de la CSG n'était pas unique (distinction revenus du patrimoine et revenus de placement)
        et il y a aussi un problème pour les années postérieures à 2017/2018
        '''
        assiette_csg_revenus_capital = foyer_fiscal('assiette_csg_revenus_capital', period)
        csg = parameters(period).taxation_capital.prelevements_sociaux.csg

        # Pour les revenus du patrimoine, le changement de CSG se fait à partir des revenus de 2017,
        # mais le taux de CSG déductible se fait à partir des revenus 2018. Pour les revenus de placement le timing est différent,
        # et reste à être pris en compte ici : cf. II.B de l'art. 67 de loi 2017-1837 et 3° et 4° du V.A de l'art. 8 de loi 2017-1836
        return -assiette_csg_revenus_capital * csg.taux_global.produits_de_placement


class crds_revenus_capital(Variable):
    value_type = float
    entity = FoyerFiscal
    label = 'CRDS sur les revenus du capital'
    definition_period = YEAR

    def formula(foyer_fiscal, period, parameters):
        assiette_csg_revenus_capital = foyer_fiscal('assiette_csg_revenus_capital', period)
        P = parameters(period).taxation_capital.prelevements_sociaux

        return -assiette_csg_revenus_capital * P.crds.produits_de_placement


class prelevements_sociaux_revenus_capital_hors_csg_crds(Variable):
    value_type = float
    entity = FoyerFiscal
    label = 'Prélèvements sociaux (hors CSG et CRDS) sur les revenus du capital'
    reference = 'https://www.service-public.fr/particuliers/vosdroits/F2329'
    definition_period = YEAR

    def formula(foyer_fiscal, period, parameters):
        assiette_csg_revenus_capital = foyer_fiscal('assiette_csg_revenus_capital', period)
        P = parameters(period).taxation_capital.prelevements_sociaux

        total = (
            P.prelevement_social.produits_de_placement
            + P.caps.produits_de_placement
            + P.prelevements_solidarite.produits_de_placement
            )

        return -assiette_csg_revenus_capital * total

    def formula_2009_01_01(foyer_fiscal, period, parameters):
        assiette_csg_revenus_capital = foyer_fiscal('assiette_csg_revenus_capital', period)
        P = parameters(period).taxation_capital.prelevements_sociaux

        total = (
            P.prelevement_social.produits_de_placement
            + P.caps.produits_de_placement
            + P.prelevements_solidarite.produits_de_placement
            + P.caps.rsa
            )

        return -assiette_csg_revenus_capital * total

    def formula_2013_01_01(foyer_fiscal, period, parameters):
        assiette_csg_revenus_capital = foyer_fiscal('assiette_csg_revenus_capital', period)
        P = parameters(period).taxation_capital.prelevements_sociaux

        total = (
            P.prelevement_social.produits_de_placement
            + P.caps.produits_de_placement
            + P.prelevements_solidarite.produits_de_placement
            )

        return -assiette_csg_revenus_capital * total

    def formula_2019_01_01(foyer_fiscal, period, parameters):
        assiette_csg_revenus_capital = foyer_fiscal('assiette_csg_revenus_capital', period)
        P = parameters(period).taxation_capital.prelevements_sociaux

        total = P.prelevements_solidarite.produits_de_placement

        return -assiette_csg_revenus_capital * total


class prelevements_sociaux_revenus_capital(Variable):
    value_type = float
    entity = FoyerFiscal
    label = 'Prélèvements sociaux sur les revenus du capital'
    reference = 'https://www.service-public.fr/particuliers/vosdroits/F2329'
    definition_period = YEAR

    def formula(foyer_fiscal, period, parameters):
        '''
        Attention : Pour les années avant 2013, cette formule n'est pas entièrement correcte car le taux de la CSG n'était pas unique (distinction revenus du patrimoine et revenus de placement)
        '''
        csg_revenus_capital = foyer_fiscal('csg_revenus_capital', period)
        crds_revenus_capital = foyer_fiscal('crds_revenus_capital', period)
        prelevements_sociaux_revenus_capital_hors_csg_crds = foyer_fiscal('prelevements_sociaux_revenus_capital_hors_csg_crds', period)

        return csg_revenus_capital + crds_revenus_capital + prelevements_sociaux_revenus_capital_hors_csg_crds



from numpy import absolute as abs_, logical_or as or_, logical_not as not_

from openfisca_france.model.base import (
    Variable,
    Individu,
    Famille,
    MONTH,
    ADD,
    TypesStatutOccupationLogement,
    TypesActivite,
    set_input_divide_by_period,
    )


class css_cmu_base_ressources_individu(Variable):
    value_type = float
    label = "Base de ressources de l'individu prise en compte pour l'éligibilité à la ACS / CMU-C / CSS"
    reference = [
        'Article R861-8 du code de la Sécurité Sociale',
        'https://www.legifrance.gouv.fr/affichCodeArticle.do?idArticle=LEGIARTI000034424885&cidTexte=LEGITEXT000006073189&dateTexte=20180829',
        'Article R861-10 du code de la Sécurité Sociale pour les ressources exclues',
        'https://www.legifrance.gouv.fr/affichCodeArticle.do?idArticle=LEGIARTI000030055485&cidTexte=LEGITEXT000006073189&dateTexte=20180829',
        "Circulaire N°DSS/2A/2002/110 du 22 février 2002 relative à la notion de ressources à prendre en compte pour l'appréciation du droit à la protection complémentaire en matière de santé",
        'http://circulaire.legifrance.gouv.fr/pdf/2009/04/cir_6430.pdf'
        ]
    entity = Individu
    definition_period = MONTH
    set_input = set_input_divide_by_period

    def formula(individu, period, parameters):
        # Rolling year
        previous_year = period.start.period('year').offset(-1)
        # N-1
        last_year = period.last_year
        last_month = period.last_month

        P = parameters(period).prestations_sociales.solidarite_insertion.minima_sociaux.cs.cmu

        ressources_a_inclure = [
            'aah',
            'allocation_securisation_professionnelle',
            'asi',
            'ass',
            'bourse_recherche',
            'caah',
            'chomage_net',
            'dedommagement_victime_amiante',
            'gains_exceptionnels',
            'indemnites_chomage_partiel',
            'indemnites_journalieres',
            'indemnites_stage',
            'indemnites_compensatrices_conges_payes',
            'pensions_alimentaires_percues',
            'pensions_invalidite',
            'prestation_compensatoire',
            'prime_forfaitaire_mensuelle_reprise_activite',
            'retraite_combattant',
            'retraite_nette',
            'revenus_stage_formation_pro',
            'rsa_base_ressources_patrimoine_individu',
            'salaire_net',
            'rente_accident_travail',
            ]

        boursier = individu('boursier', period)
        bourse = not_(boursier) * individu('bourse_enseignement_sup', period)

        ressources = sum([
            individu(ressource, previous_year, options = [ADD])
            for ressource in ressources_a_inclure
            ])

        pensions_alim_versees = abs_(individu(
            'pensions_alimentaires_versees_individu',
            previous_year, options = [ADD])
            )

        return (
            ressources
            + bourse
            + revenus_tns(individu, previous_year, last_year)
            - pensions_alim_versees
            - abbattement_chomage(individu, period, previous_year, P)
            - neutralisation_stage_formation_pro(individu, previous_year, last_month)
            )


class css_cmu_base_ressources(Variable):
    value_type = float
    label = "Base de ressources prise en compte pour l'éligibilité à la ACS / CMU-C / CSS"
    entity = Famille
    definition_period = MONTH
    set_input = set_input_divide_by_period

    def formula(famille, period, parameters):
        previous_year = period.start.period('year').offset(-1)

        ressources_famille_a_inclure = [
            'af',
            'asf',
            'aspa',
            'cf',
            'paje_clca',
            'paje_prepare',
            ]

        ressources_famille = sum([
            famille(ressource, previous_year, options = [ADD])
            for ressource in ressources_famille_a_inclure
            ])

        statut_occupation_logement = famille.demandeur.menage('statut_occupation_logement', period)
        css_cmu_forfait_logement_base = famille('css_cmu_forfait_logement_base', period)
        css_cmu_forfait_logement_al = famille('css_cmu_forfait_logement_al', period)

        P = parameters(period).prestations_sociales.solidarite_insertion.minima_sociaux.cs.cmu

        proprietaire = (statut_occupation_logement == TypesStatutOccupationLogement.proprietaire)
        heberge_titre_gratuit = (statut_occupation_logement == TypesStatutOccupationLogement.loge_gratuitement)

        forfait_logement = (
            (proprietaire + heberge_titre_gratuit)
            * css_cmu_forfait_logement_base
            + css_cmu_forfait_logement_al
            )

        ressources_individuelles = famille.members('css_cmu_base_ressources_individu', period)
        ressources_parents = famille.sum(ressources_individuelles, role = Famille.PARENT)

        age = famille.members('age', period)
        condition_enfant_a_charge = (age >= 0) * (age <= P.age_limite_pac)
        ressources_enfants = famille.sum(ressources_individuelles * condition_enfant_a_charge, role = Famille.ENFANT)

        return forfait_logement + ressources_famille + ressources_parents + ressources_enfants

# Helper functions


# Abattement sur revenus d'activité si :
# - IJ maladie
# - chômage
# - ass
# - formation professionnelle
def abbattement_chomage(individu, period, previous_year, P):
    indemnites_journalieres_maladie = individu('indemnites_journalieres_maladie', period)

    chomage = individu('activite', period) == TypesActivite.chomeur
    indemnites_chomage_partiel = individu('indemnites_chomage_partiel', period)
    chomage_net = individu('chomage_net', period)

    ass = individu('ass', period)

    revenus_stage_formation_pro = individu('revenus_stage_formation_pro', period)

    condition_ij_maladie = indemnites_journalieres_maladie > 0
    condition_chomage = chomage * ((indemnites_chomage_partiel + chomage_net) > 0)
    condition_ass = ass > 0
    condition_revenus_formation_pro = revenus_stage_formation_pro > 0

    eligibilite_abattement_chomage = or_(condition_ij_maladie, or_(condition_chomage, or_(condition_ass, condition_revenus_formation_pro)))

    salaire_net = individu('salaire_net', previous_year, options = [ADD])

    return eligibilite_abattement_chomage * salaire_net * P.abattement_chomage


# Revenus de stage de formation professionnelle exclus si plus perçus depuis 1 mois
def neutralisation_stage_formation_pro(individu, previous_year, last_month):
    revenus_stage_formation_pro_last_month = individu('revenus_stage_formation_pro', last_month)
    revenus_stage_formation_pro_annee = individu('revenus_stage_formation_pro', previous_year, options = [ADD])
    return (revenus_stage_formation_pro_last_month == 0) * revenus_stage_formation_pro_annee


def revenus_tns(individu, previous_year, last_year):
    revenus_auto_entrepreneur = individu('rpns_auto_entrepreneur_benefice', previous_year, options = [ADD])

    # Les revenus TNS hors AE sont estimés en se basant sur N-1
    rpns_micro_entreprise_benefice = individu('rpns_micro_entreprise_benefice', last_year)
    rpns_benefice_exploitant_agricole = individu('rpns_benefice_exploitant_agricole', last_year)
    rpns_autres_revenus = individu('rpns_autres_revenus', last_year)

    return revenus_auto_entrepreneur + rpns_micro_entreprise_benefice + rpns_benefice_exploitant_agricole + rpns_autres_revenus



from openfisca_france.model.base import *


class gratuite_musees_monuments(Variable):
    value_type = bool
    label = 'Accès gratuit aux musées et monuments nationaux'
    entity = Individu
    definition_period = MONTH
    set_input = set_input_dispatch_by_period
    reference = 'https://www.service-public.fr/particuliers/vosdroits/F20348'

    def formula(individu, period, parameters):
        age = individu('age', period)
        condition_age = (age >= parameters(period).prestations_sociales.gratuite_musees_monuments.age_min) * (age <= parameters(period).prestations_sociales.gratuite_musees_monuments.age_max)

        condition_nationalite = individu('resident_ue', period)

        return condition_age * condition_nationalite



from numpy import abs as abs_, logical_or as or_

from openfisca_france.model.base import *


class inapte_travail(Variable):
    value_type = bool
    entity = Individu
    label = 'Reconnu inapte au travail'
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class asi_aspa_base_ressources_individu(Variable):
    value_type = float
    label = 'Base ressources individuelle du minimum vieillesse/ASPA'
    entity = Individu
    definition_period = MONTH
    set_input = set_input_divide_by_period

    def formula(individu, period, parameters):
        last_year = period.last_year
        three_previous_months = period.last_3_months
        law = parameters(period)
        leg_1er_janvier = parameters(period.start.offset('first-of', 'year'))

        ressources_incluses = [
            'allocation_securisation_professionnelle',
            'chomage_net',
            'dedommagement_victime_amiante',
            'gains_exceptionnels',
            'indemnites_chomage_partiel',
            'indemnites_journalieres',
            'indemnites_volontariat',
            'pensions_alimentaires_percues',
            'pensions_invalidite',
            'prestation_compensatoire',
            'prime_forfaitaire_mensuelle_reprise_activite',
            'retraite_brute',
            'revenus_stage_formation_pro',
            'rsa_base_ressources_patrimoine_individu',
            'salaire_de_base',
            'traitement_indiciaire_brut'
            ]

        # Revenus du foyer fiscal que l'on projette sur le premier invidividu
        rente_viagere_titre_onereux_foyer_fiscal = individu.foyer_fiscal('rente_viagere_titre_onereux', three_previous_months, options = [ADD])
        revenus_foyer_fiscal_individu = rente_viagere_titre_onereux_foyer_fiscal * individu.has_role(FoyerFiscal.DECLARANT_PRINCIPAL)
        plus_values = individu.foyer_fiscal('assiette_csg_plus_values', period.this_year) * individu.has_role(FoyerFiscal.DECLARANT_PRINCIPAL) * (3 / 12)

        def revenus_tns():
            revenus_auto_entrepreneur = individu('rpns_auto_entrepreneur_benefice', three_previous_months, options = [ADD])
            # Les revenus TNS hors AE sont estimés en se basant sur le revenu N-1

            rpns_micro_entreprise_benefice = individu('rpns_micro_entreprise_benefice', last_year) * (3 / 12)
            rpns_benefice_exploitant_agricole = individu('rpns_benefice_exploitant_agricole', last_year) * (3 / 12)
            rpns_micro_entreprise_beneficens_autres_revenus = individu('rpns_autres_revenus', last_year) * (3 / 12)

            return (
                revenus_auto_entrepreneur
                + rpns_micro_entreprise_benefice
                + rpns_benefice_exploitant_agricole
                + rpns_micro_entreprise_beneficens_autres_revenus
                )

        pension_invalidite = (individu('pensions_invalidite', period) > 0)
        aspa_eligibilite = individu('aspa_eligibilite', period)
        asi_eligibilite = individu('asi_eligibilite', period)

        # Exclut l'AAH si éligible ASPA, retraite ou pension invalidité
        # en application du II.B. de http://www.legislation.cnav.fr/Pages/texte.aspx?Nom=LE_MIN_19031982
        aah = individu('aah', three_previous_months, options = [ADD])
        aah = aah * not_(aspa_eligibilite) * not_(asi_eligibilite) * not_(pension_invalidite)

        pensions_alimentaires_versees = individu(
            'pensions_alimentaires_versees_individu', three_previous_months, options = [ADD]
            )

        def abattement_salaire():
            aspa_couple = individu.famille('aspa_couple', period)

            # Abattement sur les salaires (appliqué sur une base trimestrielle)
            abattement_forfaitaire_base = (
                leg_1er_janvier.marche_travail.salaire_minimum.smic.smic_b_horaire * law.marche_travail.salaire_minimum.smic.nb_heures_travail_mensuel
                )

            taux_abattement_forfaitaire = where(
                aspa_couple,
                law.prestations_sociales.solidarite_insertion.minimum_vieillesse.aspa.abattement_forfaitaire.couples,
                law.prestations_sociales.solidarite_insertion.minimum_vieillesse.aspa.abattement_forfaitaire.personnes_seules
                )

            abattement_forfaitaire = abattement_forfaitaire_base * taux_abattement_forfaitaire
            salaire_de_base = individu('salaire_de_base', three_previous_months, options = [ADD])
            traitement_indiciaire_brut = individu('traitement_indiciaire_brut', three_previous_months, options = [ADD])
            return min_(salaire_de_base + traitement_indiciaire_brut, abattement_forfaitaire)

        base_ressources_3_mois = sum(
            max_(0, individu(ressource_type, three_previous_months, options = [ADD]))
            for ressource_type in ressources_incluses
            ) + aah + revenus_foyer_fiscal_individu + revenus_tns() - abs_(pensions_alimentaires_versees) - abattement_salaire() + plus_values

        return base_ressources_3_mois / 3


class asi_aspa_base_ressources(Variable):
    value_type = float
    label = 'Base ressource du minimum vieillesse et assimilés (ASPA)'
    entity = Famille
    definition_period = MONTH
    set_input = set_input_divide_by_period

    def formula(famille, period):
        base_ressources_i = famille.members('asi_aspa_base_ressources_individu', period)
        ass_i = famille.members('ass', period)
        return famille.sum(base_ressources_i + ass_i, role = Famille.PARENT)


class aspa_eligibilite(Variable):
    value_type = bool
    label = "Indicatrice individuelle d'éligibilité à l'allocation de solidarité aux personnes agées"
    entity = Individu
    definition_period = MONTH
    set_input = set_input_dispatch_by_period

    def formula(individu, period, parameters):
        age = individu('age', period)
        inapte_travail = individu('inapte_travail', period)
        taux_incapacite = individu('taux_incapacite', period)
        aah = parameters(period).prestations_sociales.prestations_etat_de_sante.invalidite.aah
        aspa = parameters(period).prestations_sociales.solidarite_insertion.minimum_vieillesse.aspa

        condition_invalidite = (taux_incapacite > aspa.taux_incapacite_aspa_anticipe) + inapte_travail
        condition_age_base = (age >= aspa.age_min)
        condition_age_anticipe = (age >= aah.age_legal_retraite) * condition_invalidite
        condition_age = condition_age_base + condition_age_anticipe
        condition_nationalite = individu('asi_aspa_condition_nationalite', period)

        return condition_age * condition_nationalite


class asi_eligibilite(Variable):
    value_type = bool
    label = "Indicatrice individuelle d'éligibilité à l'allocation supplémentaire d'invalidité"
    entity = Individu
    definition_period = MONTH
    set_input = set_input_dispatch_by_period

    def formula(individu, period):
        last_month = period.start.period('month').offset(-1)

        non_eligible_aspa = not_(individu('aspa_eligibilite', period))
        touche_pension_invalidite = individu('pensions_invalidite', period) > 0
        handicap = individu('handicap', period)
        touche_retraite = individu('retraite_nette', last_month) > 0
        condition_nationalite = individu('asi_aspa_condition_nationalite', period)

        eligible = (
            non_eligible_aspa
            * condition_nationalite
            * (handicap * touche_retraite + touche_pension_invalidite)
            )

        return eligible


class asi_aspa_condition_nationalite(Variable):
    value_type = bool
    default_value = True
    label = "Condition de nationalité et de titre de séjour pour bénéficier de l'ASPA ou l'ASI"
    entity = Individu
    definition_period = MONTH
    set_input = set_input_dispatch_by_period

    def formula(individu, period, parameters):
        ressortissant_eee = individu('ressortissant_eee', period)
        ressortissant_suisse = individu('nationalite', period) == b'CH'
        duree_possession_titre_sejour = individu('duree_possession_titre_sejour', period)
        duree_min_titre_sejour = parameters(period).prestations_sociales.solidarite_insertion.minimum_vieillesse.aspa.duree_min_titre_sejour

        return or_(ressortissant_eee, ressortissant_suisse, duree_possession_titre_sejour >= duree_min_titre_sejour)


class asi_aspa_nb_alloc(Variable):
    value_type = int
    label = "Nombre d'allocataires ASI/ASPA"
    entity = Famille
    definition_period = MONTH
    set_input = set_input_dispatch_by_period

    def formula(famille, period, parameters):
        aspa_elig_i = famille.members('aspa_eligibilite', period)
        asi_elig_i = famille.members('asi_eligibilite', period)

        nb_allocataire_asi = famille.sum(asi_elig_i, role = Famille.PARENT)
        nb_allocataire_aspa = famille.sum(aspa_elig_i, role = Famille.PARENT)

        return nb_allocataire_asi + nb_allocataire_aspa


class asi(Variable):
    value_type = float
    label = "Allocation supplémentaire d'invalidité (ASI)"
    entity = Individu
    definition_period = MONTH
    reference = 'https://www.legifrance.gouv.fr/codes/id/LEGISCTA000006156277/'
    set_input = set_input_divide_by_period
    calculate_output = calculate_output_add

    def formula_2020_04_01(individu, period, parameters):
        maries = individu.famille('maries', period)
        en_couple = individu.famille('en_couple', period)
        asi_aspa_nb_alloc = individu.famille('asi_aspa_nb_alloc', period)
        base_ressources = individu.famille('asi_aspa_base_ressources', period)
        asi = parameters(period).prestations_sociales.prestations_etat_de_sante.invalidite.asi
        aspa = parameters(period).prestations_sociales.solidarite_insertion.minimum_vieillesse.aspa

        demandeur_eligible_asi = individu.famille.demandeur('asi_eligibilite', period)
        demandeur_eligible_aspa = individu.famille.demandeur('aspa_eligibilite', period)
        conjoint_eligible_asi = individu.famille.conjoint('asi_eligibilite', period)
        conjoint_eligible_aspa = individu.famille.conjoint('aspa_eligibilite', period)

        # Un seul éligible
        elig1 = ((asi_aspa_nb_alloc == 1) & (demandeur_eligible_asi | conjoint_eligible_asi))
        # Couple d'éligibles mariés
        elig2 = demandeur_eligible_asi & conjoint_eligible_asi & maries
        # Couple d'éligibles non mariés
        elig3 = demandeur_eligible_asi & conjoint_eligible_asi & not_(maries)
        # Un seul éligible et époux éligible ASPA
        elig4 = ((demandeur_eligible_asi & conjoint_eligible_aspa) | (conjoint_eligible_asi & demandeur_eligible_aspa)) & maries
        # Un seul éligible et conjoint non marié éligible ASPA
        elig5 = ((demandeur_eligible_asi & conjoint_eligible_aspa) | (conjoint_eligible_asi & demandeur_eligible_aspa)) & not_(maries)

        plafond_ressources = (
            elig1 * (asi.plafond_ressource_seul * not_(en_couple) + asi.plafond_ressource_couple * en_couple)
            + elig2 * asi.plafond_ressource_couple
            + elig3 * asi.plafond_ressource_couple
            + elig4 * aspa.plafond_ressources.couples
            + elig5 * aspa.plafond_ressources.couples) / 12

        montant_servi_asi = max_(plafond_ressources - base_ressources, 0)

        return montant_servi_asi * (
            + individu.has_role(Famille.DEMANDEUR) * demandeur_eligible_asi * (elig1 + elig2 / 2 + elig3 / 2)
            + individu.has_role(Famille.CONJOINT) * conjoint_eligible_asi * (elig1 + elig2 / 2 + elig3 / 2)
            )

    def formula_2007(individu, period, parameters):
        maries = individu.famille('maries', period)
        en_couple = individu.famille('en_couple', period)
        asi_aspa_nb_alloc = individu.famille('asi_aspa_nb_alloc', period)
        base_ressources = individu.famille('asi_aspa_base_ressources', period)
        asi = parameters(period).prestations_sociales.prestations_etat_de_sante.invalidite.asi
        aspa = parameters(period).prestations_sociales.solidarite_insertion.minimum_vieillesse.aspa

        demandeur_eligible_asi = individu.famille.demandeur('asi_eligibilite', period)
        demandeur_eligible_aspa = individu.famille.demandeur('aspa_eligibilite', period)
        conjoint_eligible_asi = individu.famille.conjoint('asi_eligibilite', period)
        conjoint_eligible_aspa = individu.famille.conjoint('aspa_eligibilite', period)

        # Un seul éligible
        elig1 = ((asi_aspa_nb_alloc == 1) & (demandeur_eligible_asi | conjoint_eligible_asi))
        # Couple d'éligibles mariés
        elig2 = demandeur_eligible_asi & conjoint_eligible_asi & maries
        # Couple d'éligibles non mariés
        elig3 = demandeur_eligible_asi & conjoint_eligible_asi & not_(maries)
        # Un seul éligible et époux éligible ASPA
        elig4 = ((demandeur_eligible_asi & conjoint_eligible_aspa) | (conjoint_eligible_asi & demandeur_eligible_aspa)) & maries
        # Un seul éligible et conjoint non marié éligible ASPA
        elig5 = ((demandeur_eligible_asi & conjoint_eligible_aspa) | (conjoint_eligible_asi & demandeur_eligible_aspa)) & not_(maries)

        montant_max = (
            elig1 * asi.montant_seul
            + elig2 * asi.montant_couple
            + elig3 * 2 * asi.montant_seul
            + elig4 * (asi.montant_couple / 2 + aspa.montant_maximum_annuel.couples / 2)
            + elig5 * (asi.montant_seul + aspa.montant_maximum_annuel.couples / 2)) / 12

        ressources = base_ressources + montant_max

        plafond_ressources = (
            elig1 * (asi.plafond_ressource_seul * not_(en_couple) + asi.plafond_ressource_couple * en_couple)
            + elig2 * asi.plafond_ressource_couple
            + elig3 * asi.plafond_ressource_couple
            + elig4 * aspa.plafond_ressources.couples
            + elig5 * aspa.plafond_ressources.couples) / 12

        depassement = max_(ressources - plafond_ressources, 0)

        diff = (
            (elig1 | elig2 | elig3) * (montant_max - depassement)
            + elig4 * (asi.montant_couple / 12 / 2 - depassement / 2)
            + elig5 * (asi.montant_seul / 12 - depassement / 2)
            )

        # Montant mensuel servi (sous réserve d'éligibilité)
        montant_servi_asi = max_(diff, 0)
        return montant_servi_asi * (
            individu.has_role(Famille.DEMANDEUR) * demandeur_eligible_asi * (elig1 + elig2 / 2 + elig3 / 2)
            + individu.has_role(Famille.CONJOINT) * conjoint_eligible_asi * (elig1 + elig2 / 2 + elig3 / 2)
            )


class aspa_couple(Variable):
    value_type = bool
    label = "Couple au sens de l'ASPA"
    entity = Famille
    definition_period = MONTH
    set_input = set_input_dispatch_by_period

    def formula_2002_01_01(famille, period):
        maries = famille('maries', period)
        return maries

    def formula_2007_01_01(famille, period):
        en_couple = famille('en_couple', period)
        return en_couple


class aspa(Variable):
    calculate_output = calculate_output_add
    value_type = float
    entity = Famille
    label = 'Allocation de solidarité aux personnes agées'
    reference = 'http://vosdroits.service-public.fr/particuliers/F16871.xhtml'
    definition_period = MONTH
    set_input = set_input_divide_by_period

    def formula_2020_04_01(famille, period, parameters):
        maries = famille('maries', period)
        en_couple = famille('en_couple', period)
        asi_aspa_nb_alloc = famille('asi_aspa_nb_alloc', period)
        base_ressources = famille('asi_aspa_base_ressources', period)
        P = parameters(period).prestations_sociales.solidarite_insertion.minimum_vieillesse

        demandeur_eligible_asi = famille.demandeur('asi_eligibilite', period)
        demandeur_eligible_aspa = famille.demandeur('aspa_eligibilite', period)
        conjoint_eligible_asi = famille.conjoint('asi_eligibilite', period)
        conjoint_eligible_aspa = famille.conjoint('aspa_eligibilite', period)

        # Un seul éligible
        elig1 = ((asi_aspa_nb_alloc == 1) & (demandeur_eligible_aspa | conjoint_eligible_aspa))
        # Couple d'éligibles
        elig2 = (demandeur_eligible_aspa & conjoint_eligible_aspa)
        # Un seul éligible et époux éligible ASI
        elig3 = ((demandeur_eligible_asi & conjoint_eligible_aspa) | (conjoint_eligible_asi & demandeur_eligible_aspa)) & maries
        # Un seul éligible et conjoint non marié éligible ASI
        elig4 = ((demandeur_eligible_asi & conjoint_eligible_aspa) | (conjoint_eligible_asi & demandeur_eligible_aspa)) & not_(maries)

        elig = elig1 | elig2 | elig3 | elig4

        montant_asi_demandeur = famille.demandeur('asi', period)
        montant_asi_conjoint = famille.conjoint('asi', period)

        montant_max = (
            elig1 * P.aspa.montant_maximum_annuel.personnes_seules
            + elig2 * P.aspa.montant_maximum_annuel.couples
            + elig3 * (montant_asi_demandeur + P.aspa.montant_maximum_annuel.couples / 2)
            + elig4 * (montant_asi_conjoint + P.aspa.montant_maximum_annuel.couples / 2)
            ) / 12

        ressources = base_ressources + montant_max

        plafond_ressources = (
            elig1
            * (P.aspa.plafond_ressources.personnes_seules * not_(en_couple) + P.aspa.plafond_ressources.couples * en_couple)
            + (elig2 | elig3 | elig4)
            * P.aspa.plafond_ressources.couples
            ) / 12

        depassement = max_(ressources - plafond_ressources, 0)

        diff = (
            (elig1 | elig2) * (montant_max - depassement)
            + (elig3 | elig4) * (P.aspa.montant_maximum_annuel.couples / 12 / 2 - depassement / 2)
            )

        # Montant mensuel servi (sous réserve d'éligibilité)
        montant_servi_aspa = max_(diff, 0)

        # TODO: Faute de mieux, on verse l'aspa à la famille plutôt qu'aux individus
        # aspa[CHEF] = demandeur_eligible_aspa*montant_servi_aspa*(elig1 + elig2/2)
        # aspa[PART] = conjoint_eligible_aspa*montant_servi_aspa*(elig1 + elig2/2)
        return elig * montant_servi_aspa

    def formula_2006_01_01(famille, period, parameters):
        maries = famille('maries', period)
        en_couple = famille('en_couple', period)
        asi_aspa_nb_alloc = famille('asi_aspa_nb_alloc', period)
        base_ressources = famille('asi_aspa_base_ressources', period)
        asi = parameters(period).prestations_sociales.prestations_etat_de_sante.invalidite.asi
        aspa = parameters(period).prestations_sociales.solidarite_insertion.minimum_vieillesse.aspa

        demandeur_eligible_asi = famille.demandeur('asi_eligibilite', period)
        demandeur_eligible_aspa = famille.demandeur('aspa_eligibilite', period)
        conjoint_eligible_asi = famille.conjoint('asi_eligibilite', period)
        conjoint_eligible_aspa = famille.conjoint('aspa_eligibilite', period)

        # Un seul éligible
        elig1 = ((asi_aspa_nb_alloc == 1) & (demandeur_eligible_aspa | conjoint_eligible_aspa))
        # Couple d'éligibles
        elig2 = (demandeur_eligible_aspa & conjoint_eligible_aspa)
        # Un seul éligible et époux éligible ASI
        elig3 = ((demandeur_eligible_asi & conjoint_eligible_aspa) | (conjoint_eligible_asi & demandeur_eligible_aspa)) & maries
        # Un seul éligible et conjoint non marié éligible ASI
        elig4 = ((demandeur_eligible_asi & conjoint_eligible_aspa) | (conjoint_eligible_asi & demandeur_eligible_aspa)) & not_(maries)

        elig = elig1 | elig2 | elig3 | elig4

        montant_max = (
            elig1 * aspa.montant_maximum_annuel.personnes_seules
            + elig2 * aspa.montant_maximum_annuel.couples
            + elig3 * (asi.montant_couple / 2 + aspa.montant_maximum_annuel.couples / 2)
            + elig4 * (asi.montant_seul + aspa.montant_maximum_annuel.couples / 2)
            ) / 12

        ressources = base_ressources + montant_max

        plafond_ressources = (
            elig1
            * (aspa.plafond_ressources.personnes_seules * not_(en_couple) + aspa.plafond_ressources.couples * en_couple)
            + (elig2 | elig3 | elig4)
            * aspa.plafond_ressources.couples
            ) / 12

        depassement = max_(ressources - plafond_ressources, 0)

        diff = (
            (elig1 | elig2) * (montant_max - depassement)
            + (elig3 | elig4) * (aspa.montant_maximum_annuel.couples / 12 / 2 - depassement / 2)
            )

        # Montant mensuel servi (sous réserve d'éligibilité)
        montant_servi_aspa = max_(diff, 0)

        # TODO: Faute de mieux, on verse l'aspa à la famille plutôt qu'aux individus
        # aspa[CHEF] = demandeur_eligible_aspa*montant_servi_aspa*(elig1 + elig2/2)
        # aspa[PART] = conjoint_eligible_aspa*montant_servi_aspa*(elig1 + elig2/2)
        return elig * montant_servi_aspa



from numpy import isin, logical_not as not_, select

from openfisca_core.indexed_enums import Enum
from openfisca_core.periods import MONTH
from openfisca_core.variables import Variable

from openfisca_france.entities import Famille, Menage
from openfisca_france.model.base import set_input_dispatch_by_period, set_input_divide_by_period

paris_communes_limitrophes = [
    b'75056',  # Paris
    b'93001',  # Bagnolet
    b'93006',  # Boulogne-Billancourt
    b'94018',  # Charenton-le-Pont
    b'92024',  # Clichy-la-Garenne
    b'94033',  # Fontenay-sous-Bois
    b'94037',  # Gentilly
    b'92040',  # Issy-les-Moulineaux
    b'94041',  # Ivry-sur-Seine
    b'94042',  # Joinville-le-Pont
    b'94043',  # Le Kremlin-Bicêtre
    b'93045',  # Les Lilas
    b'93061',  # Le Pré-Saint-Gervais
    b'92044',  # Levallois-Perret
    b'92046',  # Malakoff
    b'93048',  # Montreuil
    b'92049',  # Montrouge
    b'92051',  # Neuilly-sur-Seine
    b'94052',  # Nogent-sur-Marne
    b'93055',  # Pantin
    b'92062',  # Puteaux
    b'92064',  # Saint-Cloud
    b'93066',  # Saint-Denis
    b'94067',  # Saint-Mandé
    b'94069',  # Saint-Maurice
    b'93070',  # Saint-Ouen
    b'92073',  # Suresnes
    b'92075',  # Vanves
    b'94080',  # Vincennes
    ]


class ZoneLogementSocial(Enum):
    __order__ = 'paris_communes_limitrophes ile_de_france autres_regions'
    paris_communes_limitrophes = 'Paris et communes limitrophes'
    ile_de_france = 'Île-de-France hors Paris et communes limitrophes'
    autres_regions = 'Autres régions'


class zone_logement_social(Variable):
    value_type = Enum
    possible_values = ZoneLogementSocial
    default_value = ZoneLogementSocial.autres_regions
    entity = Menage
    definition_period = MONTH
    set_input = set_input_dispatch_by_period
    label = 'Zone logement social'

    def formula(menage, period):
        depcom = menage('depcom', period)
        in_paris_communes_limitrophes = isin(depcom, paris_communes_limitrophes)
        in_idf = menage('residence_ile_de_france', period)

        return select(
            [
                in_paris_communes_limitrophes,
                in_idf
                ],
            [
                ZoneLogementSocial.paris_communes_limitrophes,
                ZoneLogementSocial.ile_de_france,
                ],
            default = ZoneLogementSocial.autres_regions
            )


class CategorieMenageLogementSocial(Enum):
    __order__ = 'categorie_1 categorie_2 categorie_3 categorie_4 categorie_5 categorie_6'
    categorie_1 = 'Une personne seule'
    categorie_2 = "Deux personnes ne comportant aucune pers. à charge à l'exclusion des jeunes ménages"
    categorie_3 = 'Trois personnes ou une pers. seule avec une pers. à charge ou jeune ménage sans personne à charge'
    categorie_4 = 'Quatre personnes ou une pers. seule avec deux pers. à charge'
    categorie_5 = 'Cinq personnes ou une pers. seule avec trois pers. à charge'
    categorie_6 = 'Six personnes ou une pers. seule avec quatre pers. à charge'


class logement_social_categorie_menage(Variable):
    entity = Famille
    value_type = Enum
    possible_values = CategorieMenageLogementSocial
    default_value = CategorieMenageLogementSocial.categorie_1
    definition_period = MONTH
    set_input = set_input_dispatch_by_period
    label = 'Catégorie de ménage pour déterminer le plafond de ressources'
    reference = [
        "Arrêté du 29 juillet 1987 relatif aux plafonds de ressources des bénéficiaires de la législation sur les habitations à loyer modéré et des nouvelles aides de l'Etat en secteur locatif",
        'https://www.legifrance.gouv.fr/affichTexte.do?cidTexte=JORFTEXT000000294318'
        ]

    def formula(famille, period, parameters):

        nb_personnes = famille.nb_persons()
        personnes_a_charge = famille('al_nb_personnes_a_charge', period)
        personne_seule = (famille.nb_persons(Famille.PARENT) == 1) * (personnes_a_charge == 0)

        # Jeune ménage : Couple marié, concubins ou pacsés, sans personne à charge,
        # dont la somme des âges des deux conjoints est inférieure ou égale à 55 ans
        age = famille.members('age', period)
        sum_age = famille.sum(age, role = Famille.PARENT)
        jeune_menage = (not_(personne_seule) * (sum_age <= 55))

        return select(
            [
                personne_seule,
                # Deux personnes ne comportant aucune personne à charge, à l'exclusion des jeunes ménages.
                ((nb_personnes == 2) * (personnes_a_charge == 0) * not_(jeune_menage)),
                # Trois personnes, ou une personne seule avec une personne à charge, ou jeune ménage sans personne à charge.
                (personnes_a_charge == 1) + (jeune_menage * (personnes_a_charge == 0)),
                # Quatre personnes, ou une personne seule avec deux personnes à charge.
                (personnes_a_charge == 2),
                # Cinq personnes, ou une personne seule avec trois personnes à charge.
                (personnes_a_charge == 3)
                ],
            [
                CategorieMenageLogementSocial.categorie_1,
                CategorieMenageLogementSocial.categorie_2,
                CategorieMenageLogementSocial.categorie_3,
                CategorieMenageLogementSocial.categorie_4,
                CategorieMenageLogementSocial.categorie_5
                ],
            default = CategorieMenageLogementSocial.categorie_6
            )


class logement_social_plafond_ressources(Variable):
    entity = Famille
    value_type = float
    definition_period = MONTH
    set_input = set_input_divide_by_period
    label = "Plafonds de ressources des bénéficiaires de la législation sur les habitations à loyer modéré et des nouvelles aides de l'Etat en secteur locatif"
    reference = [
        "Arrêté du 22 décembre 2016 modifiant l'arrêté du 29 juillet 1987 relatif aux plafonds de ressources des bénéficiaires de la législation sur les habitations à loyer modéré et des nouvelles aides de l'Etat en secteur locatif ",
        'https://www.legifrance.gouv.fr/eli/arrete/2016/12/22/LHAL1629455A/jo/texte',
        ]

    def formula(famille, period, parameters):
        logement_social = parameters(period).prestations_sociales.aides_logement.logement_social.plai

        categorie_menage = famille('logement_social_categorie_menage', period)
        zone_logement_social = famille.demandeur.menage('zone_logement_social', period)
        personnes_a_charge = famille('al_nb_personnes_a_charge', period)

        # On détermine le nombre de personnes à charge supplémentaires au-dessus de 4
        personnes_a_charge_supplementaires = (personnes_a_charge > 4) * (personnes_a_charge - 4)

        plafond_ressources_par_categorie = logement_social.plafond_ressources.par_categorie_de_menage[categorie_menage]
        par_personne_supplementaire = logement_social.plafond_ressources.par_personne_supplementaire[zone_logement_social]

        return plafond_ressources_par_categorie[zone_logement_social] + (personnes_a_charge_supplementaires * par_personne_supplementaire)


class logement_social_eligible(Variable):
    entity = Famille
    value_type = bool
    definition_period = MONTH
    set_input = set_input_dispatch_by_period
    label = 'Logement social - Éligibilité'
    ''' Le montant des ressources à prendre en considération pour l'attribution d'un logement HLM,
        est égal à la somme des revenus fiscaux de référence de chaque personne composant le ménage au titre de l'année n-2, soit 2016 pour 2018.
        Toutefois, il est tenu compte des revenus de l'année n-1 ou des revenus des douze derniers mois, s'ils sont inférieurs d'au moins 10 % par
        rapport à ceux de l'année n-2.'''
    def formula_2017(famille, period, parameters):
        parent_majeur = famille.any(famille.members('majeur', period), role = Famille.PARENT)
        logement_social_plafond_ressources = famille('logement_social_plafond_ressources', period)
        rfr_n0 = famille.demandeur.foyer_fiscal('rfr', period.start.offset('first-of', 'year').period('year'))
        rfr_n1 = famille.demandeur.foyer_fiscal('rfr', period.last_year)
        rfr_n2 = famille.demandeur.foyer_fiscal('rfr', period.n_2)

        elig = (rfr_n2 <= logement_social_plafond_ressources) + ((rfr_n1 < 0.9 * rfr_n2) * (rfr_n1 < logement_social_plafond_ressources)) + ((rfr_n0 < 0.9 * rfr_n2) * (rfr_n0 < logement_social_plafond_ressources))

        return parent_majeur * elig



from openfisca_france.model.base import *


class reduction_loyer_solidarite_plafond_ressources(Variable):
    value_type = float
    entity = Famille
    label = 'Plafond de ressources pour le calcul de la réduction du loyer de solidarité'
    reference = [
        'https://www.anil.org/aj-reduction-loyer-solidarite-rls-apl/',
        'https://www.legifrance.gouv.fr/affichTexte.do?cidTexte=JORFTEXT000036650010&dateTexte=&categorieLien=id',
        'https://www.legifrance.gouv.fr/affichTexte.do?cidTexte=JORFTEXT000036650026&dateTexte=&categorieLien=id'
        ]
    definition_period = MONTH

    def formula(famille, period, parameters):
        rls = parameters(period).prestations_sociales.aides_logement.reduction_loyer_solidarite
        personnes_a_charge_al = famille('al_nb_personnes_a_charge', period)
        couple = famille('al_couple', period)
        zone_apl = famille.demandeur.menage('zone_apl', period)

        plafond_ressources = rls.plafond_ressources.par_zone[zone_apl]

        plafond_personne_seule = plafond_ressources.personnes_seules
        plafond_couple = plafond_ressources.couples
        plafond_1pac = plafond_ressources.famille_1pac
        plafond_2pac = plafond_ressources.famille_2pac
        plafond_3pac = plafond_ressources.famille_3pac
        plafond_4pac = plafond_ressources.famille_4pac
        plafond_5pac = plafond_ressources.famille_5pac
        plafond_6pac = plafond_ressources.famille_6pac
        plafond_famille = plafond_ressources.famille_6pac + (personnes_a_charge_al > 6) * (personnes_a_charge_al - 6) * plafond_ressources.majoration_par_pac_supp

        return select(
            [
                not_(couple) * (personnes_a_charge_al == 0),
                couple * (personnes_a_charge_al == 0),
                personnes_a_charge_al == 1,
                personnes_a_charge_al == 2,
                personnes_a_charge_al == 3,
                personnes_a_charge_al == 4,
                personnes_a_charge_al == 5,
                personnes_a_charge_al == 6,
                ],
            [
                plafond_personne_seule,
                plafond_couple,
                plafond_1pac,
                plafond_2pac,
                plafond_3pac,
                plafond_4pac,
                plafond_5pac,
                plafond_6pac
                ],
            default = plafond_famille
            )


class reduction_loyer_solidarite_montant(Variable):
    value_type = float
    entity = Famille
    label = 'Montant de la réduction du loyer de solidarité'
    reference = 'https://www.legifrance.gouv.fr/eli/arrete/2018/2/27/TERL1801551A/jo/article_2'
    definition_period = MONTH
    set_input = set_input_divide_by_period

    def formula(famille, period, parameters):
        rls = parameters(period).prestations_sociales.aides_logement.reduction_loyer_solidarite
        personnes_a_charge_al = famille('al_nb_personnes_a_charge', period)
        couple = famille('al_couple', period)
        zone_apl = famille.demandeur.menage('zone_apl', period)

        montant = rls.montant.par_zone[zone_apl]

        montant_personne_seule = montant.personnes_seules
        montant_couple = montant.couples
        montant_1pac = montant.famille_1pac
        montant_famille = montant.famille_1pac + (personnes_a_charge_al > 1) * (personnes_a_charge_al - 1) * montant.majoration_par_pac_supp

        return select(
            [
                not_(couple) * (personnes_a_charge_al == 0),
                couple * (personnes_a_charge_al == 0),
                personnes_a_charge_al == 1
                ],
            [
                montant_personne_seule,
                montant_couple,
                montant_1pac
                ],
            default = montant_famille
            )


class reduction_loyer_solidarite(Variable):
    value_type = float
    entity = Famille
    label = 'Réduction du loyer de solidarité effectivement versée'
    reference = [
        'https://www.anil.org/aj-reduction-loyer-solidarite-rls-apl/',
        'https://www.legifrance.gouv.fr/affichTexte.do?cidTexte=JORFTEXT000036650010&dateTexte=&categorieLien=id',
        'https://www.legifrance.gouv.fr/affichTexte.do?cidTexte=JORFTEXT000036650026&dateTexte=&categorieLien=id'
        ]
    definition_period = MONTH
    set_input = set_input_divide_by_period

    def formula_2018_01_01(famille, period):
        # les ressources renvoyés sont recombiné pour donner une valeur annuelle
        # necessité de diviser par 12 pour comparer au plafond mensuel
        ressources = famille('aide_logement_base_ressources', period) / 12
        plafond = famille('reduction_loyer_solidarite_plafond_ressources', period)
        statut_occupation_logement = famille.demandeur.menage('statut_occupation_logement', period)
        logement_conventionne = famille.demandeur.menage('logement_conventionne', period)
        locataire_foyer = statut_occupation_logement == TypesStatutOccupationLogement.locataire_foyer

        eligible = (ressources < plafond) * logement_conventionne * not_(locataire_foyer)
        montant = famille('reduction_loyer_solidarite_montant', period)
        return eligible * montant



from openfisca_core.periods import Instant, Period
from openfisca_france.model.base import *


class aide_merite_eligibilite(Variable):
    value_type = bool
    entity = Individu
    label = "Éligibilité à l'aide au mérite"
    definition_period = MONTH
    reference = [
        'https://www.etudiant.gouv.fr/fr/aide-au-merite-1291',
        'https://www.service-public.fr/particuliers/vosdroits/F1010',
        'https://www.legifrance.gouv.fr/codes/id/LEGISCTA000030722211/'
        ]
    set_input = set_input_dispatch_by_period
    documentation = '''
    Complémentaire à la bourse sur critères sociaux, pour les étudiants
    ayant eu mention très bien au baccalauréat.

    Non modélisé :
    L'étudiant respecte les conditions d'inscription pédagogique, d'assiduité
    et de présentation aux examens (non applicable si nouveau bachelier)
    ou redouble pour raisons médicales.
    L'étudiant éligible à la bourse sur critères sociaux et éligible à une aide au mérite
    en année universitaire N-1 et ayant réalisé un Service Civique
    au titre de cette même année, peut percevoir son aide au mérite en N.
    Un étudiant ne peut bénéficier de plus de 3 fois de l'aide au mérite.
    '''

    def formula(individu, period):

        def periode_universitaire_precedente(mois_calcul):
            nb_mois_annee_courante = mois_calcul.date.month - mois_calcul.this_year.date.month + 1

            # https://www.campusfrance.org/fr/node/2176
            if nb_mois_annee_courante < 9:
                debut_annee_courante = mois_calcul.last_year
            else:
                debut_annee_courante = mois_calcul.this_year

            rentree_an_passe = Instant((debut_annee_courante.offset(-1), 9, 1))
            periode_universitaire_precedente = str(Period((MONTH, rentree_an_passe, 10)))
            return periode_universitaire_precedente

        # l'individu intègre un établissement supérieur à la rentrée
        etudiant = individu('etudiant', period)

        bourse_criteres_sociaux = individu('bourse_criteres_sociaux', period)
        allocation_annuelle_etudiant = individu('allocation_annuelle_etudiant', period)
        condition_ressources = (bourse_criteres_sociaux + allocation_annuelle_etudiant) > 0

        mention_baccalaureat = individu('mention_baccalaureat', period)
        condition_mention = (
            (mention_baccalaureat == TypesMention.mention_tres_bien)
            + (mention_baccalaureat == TypesMention.mention_tres_bien_felicitations_jury)
            )

        # a déjà perçu l'aide l'année [universitaire] précédente
        periode_universitaire_precedente = periode_universitaire_precedente(period)
        aide_merite_eligibilite_an_dernier = individu(
            'aide_merite_eligibilite',
            periode_universitaire_precedente,
            options = [ADD]
            )

        return etudiant * condition_ressources * (condition_mention + aide_merite_eligibilite_an_dernier)


class aide_merite_montant(Variable):
    value_type = float
    entity = Individu
    label = "Montant de l'aide au mérite pour l'année universitaire"
    definition_period = MONTH
    reference = [
        'https://www.etudiant.gouv.fr/fr/aide-au-merite-1291',
        'https://www.service-public.fr/particuliers/vosdroits/F1010',
        'https://www.legifrance.gouv.fr/codes/id/LEGISCTA000030722211/'
        ]
    documentation = '''
    Aide versée en 9 mensualités.

    Non modélisé :
    Pour un baccalauréat obtenu avant 2015, quelques conditions d'attribution diffèrent
    et le montant de l'aide est de 1800€/an.
    '''

    def formula(individu, period, parameters):
        aide_merite_eligibilite = individu('aide_merite_eligibilite', period.first_month)
        return aide_merite_eligibilite * parameters(period).prestations_sociales.aides_jeunes.bourses.bourses_enseignement_superieur.aide_merite.montant_annuel



from openfisca_core.model_api import *
from openfisca_france.entities import Famille, FoyerFiscal, Individu, Menage  # noqa F401

# Enums commonly used through the legislation


class TypesNiveauDiplome(Enum):
    # répertoire national des certifications professionnelles
    # https://fr.wikipedia.org/wiki/Liste_des_diplômes_en_France#Nomenclature_des_niveaux_de_diplômes
    # https://www.legifrance.gouv.fr/jorf/id/JORFTEXT000037964754/
    __order__ = 'non_renseigne niveau_1 niveau_2 niveau_3 niveau_4 niveau_5 niveau_6 niveau_7 niveau_8'
    non_renseigne = 'Non renseigné'
    niveau_1 = 'Niveau 1 - École maternelle'
    niveau_2 = 'Niveau 2 - École élémentaire'
    niveau_3 = 'Niveau 3 - CAP, DNP, CFG'
    niveau_4 = 'Niveau 4 - Baccalauréat'
    niveau_5 = 'Niveau 5 - Bac+2 BTS, CPGE'
    niveau_6 = 'Niveau 6 - Bac+3 Licence, BUT'
    niveau_7 = 'Niveau 7 - Bac+5 Master'
    niveau_8 = 'Niveau 8 - Bac+8 Doctorat'


class TypesMention(Enum):
    # Mentions délivrées aux candidats à l'issue d'épreuves d'enseignement.
    # https://www.legifrance.gouv.fr/codes/article_lc/LEGIARTI000037875563/2021-04-02
    __order__ = 'non_renseignee mention_assez_bien mention_bien mention_tres_bien mention_tres_bien_felicitations_jury'
    non_renseignee = 'Non renseignée'
    mention_assez_bien = 'Mention assez bien'  # [12, 14[
    mention_bien = 'Mention bien'  # [14, 16[
    mention_tres_bien = 'Mention très bien'  # [16, 18[
    mention_tres_bien_felicitations_jury = 'Mention très bien avec félicitations du jury'  # 18+


class TypesActivite(Enum):
    __order__ = 'actif chomeur etudiant retraite inactif'
    actif = 'Actif occupé'
    chomeur = 'Chômeur'
    etudiant = 'Étudiant, élève'
    retraite = 'Retraité'
    inactif = 'Autre, inactif'


class TypesSecteurActivite(Enum):
    __order__ = 'non_renseigne agricole non_agricole'
    non_renseigne = 'Non renseigné'
    agricole = 'Agricole'
    non_agricole = 'Non agricole'


class TypesCategorieNonSalarie(Enum):
    __order__ = 'non_pertinent artisan commercant profession_liberale'  # Needed to preserve the enum order in Python 2
    non_pertinent = "Non pertinent (l'individu n'est pas un travailleur indépendant)"
    artisan = 'Artisan'
    commercant = 'Commerçant'
    profession_liberale = 'Profession libérale'


class TypesCategorieSalarie(Enum):
    __order__ = 'prive_non_cadre prive_cadre public_titulaire_etat public_titulaire_militaire public_titulaire_territoriale public_titulaire_hospitaliere public_non_titulaire non_pertinent'  # Needed to preserve the enum order in Python 2
    prive_non_cadre = 'Non cadre du secteur privé'
    prive_cadre = 'Cadre du secteur privé'
    public_titulaire_etat = "Titulaire de la fonction publique d'État"
    public_titulaire_militaire = 'Titulaire de la fonction publique militaire'
    public_titulaire_territoriale = 'Titulaire de la fonction publique territoriale'
    public_titulaire_hospitaliere = 'Titulaire de la fonction publique hospitalière'
    public_non_titulaire = 'Agent non-titulaire de la fonction publique'  # Les agents non titulaires, c’est-à-dire titulaires d’aucun grade de la fonction publique, peuvent être des contractuels, des vacataires, des auxiliaires, des emplois aidés…Les assistants maternels et familiaux sont eux aussi des non-titulaires.
    non_pertinent = 'Non pertinent'


class TypesStatutMarital(Enum):
    # La mention jeune_veuf indique que la personne est devenue veuve dans l'année.
    __order__ = 'non_renseigne marie celibataire divorce veuf pacse jeune_veuf'  # Needed to preserve the enum order in Python 2
    non_renseigne = 'Non renseigné'
    marie = 'Marié'
    celibataire = 'Célibataire'
    divorce = 'Divorcé'
    veuf = 'Veuf'
    pacse = 'Pacsé'
    jeune_veuf = 'Veuf d’un conjoint décédé dans l’année'


class TypesStatutOccupationLogement(Enum):
    __order__ = 'non_renseigne primo_accedant proprietaire locataire_hlm locataire_vide locataire_meuble loge_gratuitement locataire_foyer sans_domicile'  # Needed to preserve the enum order in Python 2
    non_renseigne = 'Non renseigné'
    primo_accedant = 'Accédant à la propriété'
    proprietaire = 'Propriétaire (non accédant) du logement'
    locataire_hlm = "Locataire d'un logement HLM"
    locataire_vide = "Locataire ou sous-locataire d'un logement loué vide non-HLM"
    locataire_meuble = "Locataire ou sous-locataire d'un logement loué meublé ou d'une chambre d'hôtel"
    loge_gratuitement = "Logé gratuitement par des parents, des amis ou l'employeur"
    locataire_foyer = "Locataire d'un foyer (résidence universitaire, maison de retraite, foyer de jeune travailleur, résidence sociale...)"
    sans_domicile = 'Sans domicile stable'


# Taux de prime moyen de la fonction publique
TAUX_DE_PRIME = 0.195  # primes_fonction_publique (hors suppl. familial et indemnité de résidence)/rémunération brute



from openfisca_france.model.base import *


# Rentes viagères
class f1aw(Variable):
    cerfa_field = '1AW'
    value_type = int
    unit = 'currency'
    entity = FoyerFiscal
    label = "Rentes viagères à titre onéreux perçues par le foyer par âge d'entrée en jouissance : Moins de 50 ans"
    definition_period = YEAR


class f1bw(Variable):
    cerfa_field = '1BW'
    value_type = int
    unit = 'currency'
    entity = FoyerFiscal
    label = "Rentes viagères à titre onéreux perçues par le foyer par âge d'entrée en jouissance : De 50 à 59 ans"
    definition_period = YEAR


class f1cw(Variable):
    cerfa_field = '1CW'
    value_type = int
    unit = 'currency'
    entity = FoyerFiscal
    label = "Rentes viagères à titre onéreux perçues par le foyer par âge d'entrée en jouissance : De 60 à 69 ans"
    definition_period = YEAR


class f1dw(Variable):
    cerfa_field = '1DW'
    value_type = int
    unit = 'currency'
    entity = FoyerFiscal
    label = "Rentes viagères à titre onéreux perçues par le foyer par âge d'entrée en jouissance : A partir de 70 ans"
    definition_period = YEAR


# Revenus fonciers
class f4ba(Variable):
    cerfa_field = '4BA'
    value_type = int
    unit = 'currency'
    entity = FoyerFiscal
    label = 'Revenus fonciers imposables'
    definition_period = YEAR


class f4bb(Variable):
    cerfa_field = '4BB'
    value_type = int
    unit = 'currency'
    entity = FoyerFiscal
    label = 'Déficit imputable sur les revenus fonciers'
    definition_period = YEAR


class f4bc(Variable):
    cerfa_field = '4BC'
    value_type = int
    unit = 'currency'
    entity = FoyerFiscal
    label = 'Déficit imputable sur le revenu global'
    definition_period = YEAR


class f4bd(Variable):
    cerfa_field = '4BD'
    value_type = int
    unit = 'currency'
    entity = FoyerFiscal
    label = 'Déficits antérieurs non encore imputés'
    definition_period = YEAR


class f4be(Variable):
    cerfa_field = '4BE'
    value_type = int
    unit = 'currency'
    entity = FoyerFiscal
    label = 'Micro foncier: recettes brutes sans abattement'
    definition_period = YEAR


# Prime d'assurance loyers impayés
class f4bf(Variable):
    cerfa_field = '4BF'
    value_type = int
    unit = 'currency'
    entity = FoyerFiscal
    label = "Primes d'assurance pour loyers impayés des locations conventionnées"
    end = '2016-12-31'
    definition_period = YEAR


class f4bl(Variable):
    value_type = int
    entity = FoyerFiscal
    end = '2009-12-31'
    definition_period = YEAR
    # TODO: cf 2010 2011


# Variables utilisées par mes aides
class valeur_patrimoine_loue(Variable):
    value_type = float
    entity = Individu
    label = 'Valeur des biens immobiliers et des terrains loués'
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class revenus_locatifs(Variable):
    value_type = float
    entity = Individu
    label = 'Revenus locatifs'
    definition_period = MONTH
    set_input = set_input_divide_by_period

    def formula(individu, period):
        '''
        Revenus locatifs utilisés pour les bases ressources des minima sociaux
        Si cette variable n'est pas renseignée, on prend les revenus fonciers catégoriels de la section 4 de la déclaration 2042
        Néanmoins, selon la législation, le concept de revenus locatifs pour ces bases ressources peut être plus large, notamment
        pour les indépendants (exemple : case 5ND du formulaire 2042-C-PRO). Ce point reste à améliorer.
        '''
        revenu_categoriel_foncier = individu.foyer_fiscal('revenu_categoriel_foncier', period.this_year)
        montant = revenu_categoriel_foncier * individu.has_role(FoyerFiscal.DECLARANT_PRINCIPAL) / 12
        return max_(montant, 0)


class valeur_immo_non_loue(Variable):
    value_type = float
    entity = Individu
    label = 'Valeur des biens immobiliers possédés et non loués'
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class valeur_locative_immo_non_loue(Variable):
    value_type = float
    entity = Individu
    label = "Valeur locative, à l'année, des biens immobiliers possédés et non loués"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class valeur_terrains_non_loues(Variable):
    value_type = float
    entity = Individu
    label = 'Valeur des terrains possédés et non loués'
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class valeur_locative_terrains_non_loues(Variable):
    value_type = float
    entity = Individu
    label = "Valeur locative, à l'année, des terrains possédés et non loués"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period



from openfisca_france.model.base import Individu, Variable, MONTH, set_input_dispatch_by_period
from openfisca_france.model.prestations.education import TypesScolarite


class crous_repas_un_euro_eligibilite(Variable):
    value_type = bool
    label = 'Éligibilité au repas Crous à un euro'
    entity = Individu
    definition_period = MONTH
    set_input = set_input_dispatch_by_period
    reference = 'https://www.etudiant.gouv.fr/sites/default/files/2021-01/le-repas-un-euro-4977.pdf'
    documentation = '''
    Suite à la crise Covid-19, tous les étudiants, boursiers ou non, peuvent bénéficier
    de deux repas par jour au tarif de 1 euro.
    Pour en bénéficier, il suffit d’activer son compte Izly.
    Pour les étudiants déjà titulaires d’un compte Izly,
    la modification de tarif a été effectuée à partir du vendredi 22 janvier
    et jusqu’au début de la semaine du 25 janvier.
    '''

    def formula_2021_01(individu, period):
        return individu('scolarite', period) == TypesScolarite.enseignement_superieur

    def formula_2021_07(individu, period):
        enseignement_superieur = individu('scolarite', period) == TypesScolarite.enseignement_superieur
        detention_carte_des_metiers = individu('detention_carte_des_metiers', period)
        boursier = individu('boursier', period)
        return boursier * (enseignement_superieur + detention_carte_des_metiers)



import logging
from openfisca_france.model.base import *

log = logging.getLogger(__name__)

# Calcul de la partie du prélèvement forfaitaire unique (PFU) associé à l'impôt sur le revenu
# (le reste étant associé aux prélèvements sociaux)


class assurance_vie_pfu_ir_plus8ans_1990_19970926(Variable):
    value_type = float
    entity = FoyerFiscal
    label = "Produits des bons ou contrats de capitalisation et d'assurance vie d'une durée d'au moins 8 ans pour les contrats souscrits entre le 1er janvier 1990 et le 26 septembre 1997, et que le bénéficiaire décide de soumettre au prélèvement forfaitaire unique au titre de l'impôt sur le revenu"
    definition_period = YEAR


class assurance_vie_pfu_ir_plus6ans_avant1990(Variable):
    value_type = float
    entity = FoyerFiscal
    label = "Produits des bons ou contrats de capitalisation et d'assurance vie d'une durée d'au moins 6 ans pour les contrats souscrits avant le 1er janvier 1990, et que le bénéficiaire décide de soumettre au prélèvement forfaitaire unique au titre de l'impôt sur le revenu"
    reference = 'https://www.legifrance.gouv.fr/affichTexteArticle.do?idArticle=LEGIARTI000036377422&cidTexte=JORFTEXT000036339197'  # Art. 28 (modif art 125 A, 125-0 A, 200 A et art 117 quater du CGI)
    definition_period = YEAR


class assurance_vie_pfu_ir_moins4ans_1990_19970926(Variable):
    value_type = float
    entity = FoyerFiscal
    label = "Produits des bons ou contrats de capitalisation et d'assurance vie d'une durée de moins de 4 ans pour les contrats souscrits entre le 1er janvier 1990 et le 26 septembre 1997, et que le bénéficiaire décide de soumettre au prélèvement forfaitaire unique au titre de l'impôt sur le revenu"
    reference = 'https://www.legifrance.gouv.fr/affichTexteArticle.do?idArticle=LEGIARTI000036377422&cidTexte=JORFTEXT000036339197'  # Art. 28 (modif art 125 A, 125-0 A, 200 A et art 117 quater du CGI)
    definition_period = YEAR


class assurance_vie_pfu_ir_4_8_ans_1990_19970926(Variable):
    value_type = float
    entity = FoyerFiscal
    label = "Produits des bons ou contrats de capitalisation et d'assurance vie d'une durée entre 4 et 8 ans pour les contrats souscrits entre le 1er janvier 1990 et le 26 septembre 1997, et que le bénéficiaire décide de soumettre au prélèvement forfaitaire unique au titre de l'impôt sur le revenu"
    definition_period = YEAR


class assurance_vie_pfu_ir_plus8ans_19970926_primes_avant_20170927(Variable):
    value_type = float
    entity = FoyerFiscal
    label = "Produits des bons ou contrats de capitalisation et d'assurance vie de plus de 8 ans pour les contrats souscrits après le 26 septembre 1997, dont le produits sont associés aux primes versées avant le 27 septembre 2017, et que le bénéficiaire décide de soumettre au prélèvement forfaitaire unique au titre de l'impôt sur le revenu"
    reference = 'https://www.legifrance.gouv.fr/affichTexteArticle.do?idArticle=LEGIARTI000036377422&cidTexte=JORFTEXT000036339197'  # Art. 28 (modif art 125 A, 125-0 A, 200 A et art 117 quater du CGI)
    definition_period = YEAR


class assurance_vie_pfu_ir_4_8_ans_19970926_primes_avant_20170927(Variable):
    value_type = float
    entity = FoyerFiscal
    label = "Produits des bons ou contrats de capitalisation et d'assurance vie entre 4 et 8 ans pour les contrats souscrits après le 26 septembre 1997, dont le produits sont associés aux primes versées avant le 27 septembre 2017, et que le bénéficiaire décide de soumettre au prélèvement forfaitaire unique au titre de l'impôt sur le revenu"
    definition_period = YEAR


class assurance_vie_pfu_ir_moins4ans_19970926_primes_avant_20170927(Variable):
    value_type = float
    entity = FoyerFiscal
    label = "Produits des bons ou contrats de capitalisation et d'assurance vie de moins de 4 ans pour les contrats souscrits après le 26 septembre 1997, dont le produits sont associés aux primes versées avant le 27 septembre 2017, et que le bénéficiaire décide de soumettre au prélèvement forfaitaire unique au titre de l'impôt sur le revenu"
    reference = 'https://www.legifrance.gouv.fr/affichTexteArticle.do?idArticle=LEGIARTI000036377422&cidTexte=JORFTEXT000036339197'  # Art. 28 (modif art 125 A, 125-0 A, 200 A et art 117 quater du CGI)
    definition_period = YEAR


class f2zz(Variable):
    cerfa_field = '2ZZ'
    value_type = float
    entity = FoyerFiscal
    label = "Produits des bons ou contrats de capitalisation et d'assurance vie de moins de 8 ans pour les contrats souscrits après le 26 septembre 1997, dont le produits sont associés aux primes versées après le 27 septembre 2017, et que le bénéficiaire décide de soumettre au prélèvement forfaitaire unique au titre de l'impôt sur le revenu"
    definition_period = YEAR
    # start_date = date(2018, 1, 1)


class f2vv(Variable):
    value_type = float
    entity = FoyerFiscal
    label = "Produits des bons ou contrats de capitalisation et d'assurance vie de plus de 8 ans pour les contrats souscrits après le 26 septembre 1997, dont le produits sont associés aux primes versées après le 27 septembre 2017, et que le bénéficiaire décide de soumettre au prélèvement forfaitaire unique au titre de l'impôt sur le revenu; produit correspondant aux primes n'excédant pas 150 000 euros."
    definition_period = YEAR
    # start_date = date(2018, 1, 1)


class f2ww(Variable):
    value_type = float
    entity = FoyerFiscal
    label = "Produits des bons ou contrats de capitalisation et d'assurance vie de plus de 8 ans pour les contrats souscrits après le 26 septembre 1997, dont le produits sont associés aux primes versées après le 27 septembre 2017, et que le bénéficiaire décide de soumettre au prélèvement forfaitaire unique au titre de l'impôt sur le revenu; produit correspondant aux primes excédant pas 150 000 euros."
    definition_period = YEAR
    # start_date = date(2018, 1, 1)


class assurance_vie_pfu_ir(Variable):
    value_type = float
    entity = FoyerFiscal
    label = "Produits des bons ou contrats de capitalisation et d'assurance vie soumis au prélèvement forfaitaire unique au titre de l'impôt sur le revenu"
    definition_period = YEAR

    def formula_2018_01_01(foyer_fiscal, period):
        f2zz = foyer_fiscal('f2zz', period)
        f2vv = foyer_fiscal('f2vv', period)
        f2ww = foyer_fiscal('f2ww', period)

        return f2zz + f2vv + f2ww


class revenus_capitaux_prelevement_forfaitaire_unique_ir(Variable):
    value_type = float
    entity = FoyerFiscal
    label = 'Revenus des valeurs et capitaux mobiliers soumis au prélèvement forfaitaire unique (partie impôt sur le revenu)'
    definition_period = MONTH
    set_input = set_input_divide_by_period

    def formula_2018_01_01(foyer_fiscal, period, parameters):
        '''
        Note : cette variable est définie à l'échelle du mois pour être en cohérence avec les variables qu'elle remplace
                (à savoir revenus_capitaux_prelevement_bareme et revenus_capitaux_prelevement_liberatoire)
        '''
        year = period.this_year
        assurance_vie_pfu_ir = foyer_fiscal('assurance_vie_pfu_ir', year)
        f2dc = foyer_fiscal('f2dc', year)
        f2fu = foyer_fiscal('f2fu', year)
        f2ts = foyer_fiscal('f2ts', year)
        f2tr = foyer_fiscal('f2tr', year)
        f2tt = foyer_fiscal('f2tt', year)
        f2go = foyer_fiscal('f2go', year)

        return (assurance_vie_pfu_ir + f2dc + f2fu + f2ts + f2tr + f2tt + f2go) / 12


class plus_values_prelevement_forfaitaire_unique_ir(Variable):
    value_type = float
    entity = FoyerFiscal
    label = 'Plus-values soumises au prélèvement forfaitaire unique (partie impôt sur le revenu)'
    reference = 'https://www.legifrance.gouv.fr/loda/article_lc/LEGIARTI000036377422/'
    definition_period = YEAR

    def formula_2018_01_01(foyer_fiscal, period, parameters):
        f3sb = foyer_fiscal('f3sb', period)
        f3ua = foyer_fiscal('f3ua', period)
        f3va = foyer_fiscal('f3va', period)
        f3vg = foyer_fiscal('f3vg', period)
        f3tj = foyer_fiscal('f3tj', period)

        return f3sb + max_(0, f3ua - f3va) + f3vg + f3tj

    def formula_2019_01_01(foyer_fiscal, period, parameters):
        f3sb = foyer_fiscal('f3sb', period)
        f3ua = foyer_fiscal('f3ua', period)
        f3va = foyer_fiscal('f3va', period)
        f3vg = foyer_fiscal('f3vg', period)
        f3tj = foyer_fiscal('f3tj', period)
        f3tk = foyer_fiscal('f3tk', period)
        f3vt = foyer_fiscal('f3vt', period)
        f3an = foyer_fiscal('f3an', period)

        return f3sb + max_(0, f3ua - f3va) + f3vg + max_(0, f3tj - f3tk) + f3vt + f3an


class prelevement_forfaitaire_unique_ir_hors_assurance_vie_epargne_solidaire_etats_non_cooperatifs(Variable):
    value_type = float
    entity = FoyerFiscal
    label = "Partie du prélèvement forfaitaire unique associée à l'impôt sur le revenu (hors assurance-vie, épargne solidaire et produits venant des états non-coopératifs)"
    reference = [
        'Article 28 de la Loi n° 2017-1837 du 30 décembre 2017 de finances pour 2018 (modifie art. 125 A, 125-0 A, 200 A et art. 117 quater du CGI)',
        'https://www.legifrance.gouv.fr/affichTexteArticle.do?idArticle=LEGIARTI000036377422&cidTexte=JORFTEXT000036339197'
        ]
    definition_period = YEAR

    def formula_2018_01_01(foyer_fiscal, period, parameters):
        P = parameters(period).taxation_capital.prelevement_forfaitaire.partir_2018

        # Revenus des valeurs et capitaux mobiliers hors assurance-vie et hors produits d'épargne solidaire ou des états non-coopératifs
        #   Note : Les revenus d'assurance-vie, de l'épargne solidaire et des produits des états non-coopératifs ont été ajoutés dans les variables f2ee et f2dh (cf. docstring de ces varables pour une explication), d'où le fait qu'on soustrait ici ces variables de revenus_capitaux_prelevement_forfaitaire_unique_ir
        revenus_capitaux_prelevement_forfaitaire_unique_ir = foyer_fiscal('revenus_capitaux_prelevement_forfaitaire_unique_ir', period, options = [ADD])
        assurance_vie_pfu_ir = foyer_fiscal('assurance_vie_pfu_ir', period)
        produit_epargne_solidaire = foyer_fiscal('produit_epargne_solidaire', period)
        produit_etats_non_cooperatif = foyer_fiscal('produit_etats_non_cooperatif', period)
        revenus_capitaux_prelevement_forfaitaire_unique_ir_hors_assurance_vie_epargne_solidaire_etats_non_cooperatifs = max_(
            0,
            revenus_capitaux_prelevement_forfaitaire_unique_ir
            - assurance_vie_pfu_ir
            - produit_epargne_solidaire
            - produit_etats_non_cooperatif
            )

        # Intérêts des PEL et CEL, hors intérêts des PEL de plus de 12 ans, qui sont déclarés dans la déclaration de revenus (attention, on ne connait pas le formulaire 2019 des revenus 2018 au moment de faire ce code)
        interets_plan_epargne_logement_moins_de_12_ans_ouvert_a_partir_de_2018_i = foyer_fiscal.members('interets_plan_epargne_logement_moins_de_12_ans_ouvert_a_partir_de_2018', period)
        interets_plan_epargne_logement_moins_de_12_ans_ouvert_a_partir_de_2018 = foyer_fiscal.sum(interets_plan_epargne_logement_moins_de_12_ans_ouvert_a_partir_de_2018_i)
        interets_compte_epargne_logement_ouvert_a_partir_de_2018_i = foyer_fiscal.members('interets_compte_epargne_logement_ouvert_a_partir_de_2018', period)
        interets_compte_epargne_logement_ouvert_a_partir_de_2018 = foyer_fiscal.sum(interets_compte_epargne_logement_ouvert_a_partir_de_2018_i)

        # Plus-values
        plus_values_prelevement_forfaitaire_unique_ir = foyer_fiscal('plus_values_prelevement_forfaitaire_unique_ir', period)

        assiette_pfu_hors_assurance_vie = (
            revenus_capitaux_prelevement_forfaitaire_unique_ir_hors_assurance_vie_epargne_solidaire_etats_non_cooperatifs
            + interets_plan_epargne_logement_moins_de_12_ans_ouvert_a_partir_de_2018
            + interets_compte_epargne_logement_ouvert_a_partir_de_2018
            + plus_values_prelevement_forfaitaire_unique_ir
            )

        return -assiette_pfu_hors_assurance_vie * P.taux_prelevement_forfaitaire_rev_capital_eligibles_pfu_interets_dividendes_etc


class prelevement_forfaitaire_unique_ir_sur_assurance_vie(Variable):
    value_type = float
    entity = FoyerFiscal
    label = "Partie du prélèvement forfaitaire unique associée à l'impôt sur le revenu sur l'assurance-vie"
    definition_period = YEAR

    def formula_2018_01_01(foyer_fiscal, period, parameters):
        P1_taux = parameters(period).taxation_capital.prelevement_forfaitaire.partir_2018.taux_prelevement_forfaitaire_rev_capital_eligibles_pfu_interets_dividendes_etc
        P1_taux_reduit_av = parameters(period).taxation_capital.prelevement_forfaitaire.partir_2018.taux_prelevement_produits_assurance_vie_non_eligibles_prelevement_forfaitaire_unique
        P2 = parameters(period).impot_revenu.calcul_revenus_imposables.rvcm

        maries_ou_pacses = foyer_fiscal('maries_ou_pacses', period)
        f2ch = foyer_fiscal('f2ch', period)
        f2zz = foyer_fiscal('f2zz', period)
        f2vv = foyer_fiscal('f2vv', period)
        f2ww = foyer_fiscal('f2ww', period)

        abattement_residuel = max_(P2.produits_assurances_vies_assimiles.abattement * (1 + maries_ou_pacses) - f2ch, 0)
        abattement_residuel2 = max_(abattement_residuel - f2vv, 0)
        pfu_ir_sur_assurance_vie = -(
            (f2zz * P1_taux)
            + (max_(f2vv - abattement_residuel, 0) * P1_taux_reduit_av)
            + (max_(f2ww - abattement_residuel2, 0) * P1_taux)
            )

        return pfu_ir_sur_assurance_vie


class prelevement_forfaitaire_unique_ir(Variable):
    value_type = float
    entity = FoyerFiscal
    label = "Partie du prélèvement forfaitaire unique associée à l'impôt sur le revenu"
    definition_period = YEAR

    def formula_2018_01_01(foyer_fiscal, period, parameters):
        prelevement_forfaitaire_unique_ir_hors_assurance_vie_epargne_solidaire_etats_non_cooperatifs = foyer_fiscal('prelevement_forfaitaire_unique_ir_hors_assurance_vie_epargne_solidaire_etats_non_cooperatifs', period)
        prelevement_forfaitaire_unique_ir_sur_assurance_vie = foyer_fiscal('prelevement_forfaitaire_unique_ir_sur_assurance_vie', period)

        return (
            prelevement_forfaitaire_unique_ir_hors_assurance_vie_epargne_solidaire_etats_non_cooperatifs
            + prelevement_forfaitaire_unique_ir_sur_assurance_vie
            )
